/*
 * Copyright (c) 2003-2006
 * Distributed Systems Software.  All rights reserved.
 * See the file LICENSE for redistribution information.
 */

/*****************************************************************************
 * COPYRIGHT AND PERMISSION NOTICE
 *
 * Copyright (c) 2001-2003 The Queen in Right of Canada
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, provided that the above copyright notice(s) and this
 * permission notice appear in all copies of the Software and that both the
 * above copyright notice(s) and this permission notice appear in supporting
 * documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE
 * BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Except as contained in this notice, the name of a copyright holder shall not
 * be used in advertising or otherwise to promote the sale, use or other
 * dealings in this Software without prior written authorization of the
 * copyright holder.
 ***************************************************************************/

/*
 * Expression evaluation for ACS
 */

#ifndef lint
static const char copyright[] =
"Copyright (c) 2003-2006\n\
Distributed Systems Software.  All rights reserved.";
static const char revid[] =
  "$Id: expr.c 1215 2006-06-01 22:48:23Z brachman $";
#endif

#include "auth.h"
#include "group.h"
#include "acs.h"
#include "http.h"

#include <math.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

static char *log_module_name = "acs_expr";

/* Token classes */
typedef enum {
  TC_UNARY   = 0,	/* Unary operator */
  TC_BINARY  = 1,	/* Binary operator */
  TC_TERNARY = 2,	/* Ternary operator */
  TC_OPERAND = 3,	/* An operand */
  TC_SPEC    = 4	/* Special token */
} Token_class;

typedef struct Lex_token {
  Token token;
  char *lexeme;
  char *token_name;
  Token_class tc;
  int type_length;
  char *startaddr;
  Value *value;
} Lex_token;

typedef struct Func {
  const char *name;		/* Function name */
  int arity;			/* Number of arguments (-1 means a variable number) */
  char *argdesc;		/* Argument type descriptor string */
  int (*func)(Lex_state *, Arglist *, Expr_result *);
} Func;

#define L_CHARS(N)		(N)
#define L_ALPHA(N)		-(N)
#define IS_L_CHARS(N)	((N) > 0)
#define IS_L_ALPHA(N)	((N) < 0)

static inline int
lexeme_length(Lex_token *t)
{

  if (IS_L_CHARS(t->type_length))
	return(L_CHARS(t->type_length));
  else if (IS_L_ALPHA(t->type_length))
	return(L_ALPHA(t->type_length));

  return(-1);
}

static inline int
is_numeric_value(Value *v)
{

  return(v->token == T_INTEGER || v->token == T_REAL);
}

static inline int
is_operator(Lex_token *t)
{

  return(t->tc == TC_UNARY || t->tc == TC_BINARY || t->tc == TC_TERNARY);
}

/*
 * Entries are ordered so that we can scan the table from top to bottom.
 */
static Lex_token tokens[] = {
  { T_IF,        "if",     "T_IF",      TC_SPEC,   L_ALPHA(2), NULL, NULL },
  { T_ELSEIF,    "elseif", "T_ELSEIF",  TC_SPEC,   L_ALPHA(6), NULL, NULL },
  { T_ELSE,      "else",   "T_ELSE",    TC_SPEC,   L_ALPHA(4), NULL, NULL },
  { T_LBRACE,    "{",      "T_LBRACE",  TC_SPEC,   L_CHARS(1), NULL, NULL },
  { T_RBRACE,    "}",      "T_RBRACE",  TC_SPEC,   L_CHARS(1), NULL, NULL },

  { T_OR,        "||",   "T_OR",        TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_OR,        "or",   "T_OR",        TC_BINARY, L_ALPHA(2), NULL, NULL },
  { T_AND,       "&&",   "T_AND",       TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_AND,       "and",  "T_AND",       TC_BINARY, L_ALPHA(3), NULL, NULL },

  { T_PLUS_EQ,   "+=",   "T_PLUS_EQ",   TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_MINUS_EQ,  "-=",   "T_MINUS_EQ",  TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_TIMES_EQ,  "*=",   "T_TIMES_EQ",  TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_DIV_EQ,    "/=",   "T_DIV_EQ",    TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_MOD_EQ,    "%=",   "T_MOD_EQ",    TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_BITSHR_EQ, ">>=",  "T_BITSHR_EQ", TC_BINARY, L_CHARS(3), NULL, NULL },
  { T_BITSHL_EQ, "<<=",  "T_BITSHL_EQ", TC_BINARY, L_CHARS(3), NULL, NULL },
  { T_BITAND_EQ, "&=",   "T_BITAND_EQ", TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_BITXOR_EQ, "^=",   "T_BITXOR_EQ", TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_BITOR_EQ,  "|=",   "T_BITOR_EQ",  TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_DOT_EQ,    ".=",   "T_DOT_EQ",    TC_BINARY, L_CHARS(2), NULL, NULL },

  { T_BITSHL,    "<<",   "T_BITSHL",    TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_BITSHR,    ">>",   "T_BITSHR",    TC_BINARY, L_CHARS(2), NULL, NULL },

  { T_LE_I,      "le:i", "T_LE_I",      TC_BINARY, L_CHARS(4), NULL, NULL },
  { T_LT_I,      "lt:i", "T_LT_I",      TC_BINARY, L_CHARS(4), NULL, NULL },
  { T_EQ_I,      "eq:i", "T_EQ_I",      TC_BINARY, L_CHARS(4), NULL, NULL },
  { T_NE_I,      "ne:i", "T_NE_I",      TC_BINARY, L_CHARS(4), NULL, NULL },
  { T_GE_I,      "ge:i", "T_GE_I",      TC_BINARY, L_CHARS(4), NULL, NULL },
  { T_GT_I,      "gt:i", "T_GT_I",      TC_BINARY, L_CHARS(4), NULL, NULL },
  { T_LE,        "<=",   "T_LE",        TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_LE,        "le",   "T_LE",        TC_BINARY, L_ALPHA(2), NULL, NULL },
  { T_LT,        "<",    "T_LT",        TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_LT,        "lt",   "T_LT",        TC_BINARY, L_ALPHA(2), NULL, NULL },
  { T_GE,        ">=",   "T_GE",        TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_GE,        "ge",   "T_GE",        TC_BINARY, L_ALPHA(2), NULL, NULL },
  { T_GT,        ">",    "T_GT",        TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_GT,        "gt",   "T_GT",        TC_BINARY, L_ALPHA(2), NULL, NULL }, 
  { T_EQ,        "==",   "T_EQ",        TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_EQ,        "eq",   "T_EQ",        TC_BINARY, L_ALPHA(2), NULL, NULL }, 
  { T_NE,        "!=",   "T_NE",        TC_BINARY, L_CHARS(2), NULL, NULL },
  { T_NE,        "ne",   "T_NE",        TC_BINARY, L_ALPHA(2), NULL, NULL },

  { T_BITOR,     "|",    "T_BITOR",     TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_BITXOR,    "^",    "T_BITXOR",    TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_BITAND,    "&",    "T_BITAND",    TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_BITCPL,    "~",    "T_BITCPL",    TC_UNARY,  L_CHARS(1), NULL, NULL },
  { T_PLUS,      "+",    "T_PLUS",      TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_MINUS,     "-",    "T_MINUS",     TC_BINARY, L_CHARS(1), NULL, NULL }, 
  { T_EXP,       "**",   "T_EXP",       TC_BINARY, L_CHARS(2), NULL, NULL }, 
  { T_TIMES,     "*",    "T_TIMES",     TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_DIV,       "/",    "T_DIV",       TC_BINARY, L_CHARS(1), NULL, NULL }, 
  { T_MOD,       "%",    "T_MOD",       TC_BINARY, L_CHARS(1), NULL, NULL }, 
  { T_DOT,       ".",    "T_DOT",       TC_BINARY, L_CHARS(1), NULL, NULL },
  { T_COND,      "?",    "T_COND",      TC_SPEC,   L_CHARS(1), NULL, NULL },
  { T_COLON,     ":",    "T_COLON",     TC_SPEC,   L_CHARS(1), NULL, NULL },
  { T_COMMA,     ",",    "T_COMMA",     TC_SPEC,   L_CHARS(1), NULL, NULL }, 
  { T_NOT,       "!",    "T_NOT",       TC_UNARY,  L_CHARS(1), NULL, NULL },
  { T_NOT,       "not",  "T_NOT",       TC_UNARY,  L_ALPHA(3), NULL, NULL }, 
  { T_LPAREN,    "(",    "T_LPAREN",    TC_SPEC,   L_CHARS(1), NULL, NULL }, 
  { T_RPAREN,    ")",    "T_RPAREN",    TC_SPEC,   L_CHARS(1), NULL, NULL }, 
  { T_EOS,       ";",    "T_EOS",       TC_SPEC,   L_CHARS(1), NULL, NULL }, 
  { T_ASSIGN,    "=",    "T_ASSIGN",    TC_BINARY, L_CHARS(1), NULL, NULL }, 
  /* Numbers and strings are implicit at this point. */
  { T_UNDEF,     NULL,   "T_UNDEF",     TC_SPEC,   L_CHARS(0), NULL, NULL }
};

static Lex_token token_eoi = {
  T_EOI, NULL, "T_EOI", TC_SPEC, L_CHARS(0), NULL, NULL
};

static Lex_token token_unary_minus = {
  T_UNARY_MINUS, "-", "T_UNARY_MINUS", TC_UNARY, L_CHARS(1), NULL, NULL
};

static Lex_token token_unary_plus = {
  T_UNARY_PLUS, "+", "T_UNARY_PLUS", TC_UNARY, L_CHARS(1), NULL, NULL
};

static Lex_token token_preinc = {
  T_PREINC, "++", "T_PREINC", TC_UNARY, L_CHARS(2), NULL, NULL
};

static Lex_token token_predec = {
  T_PREDEC, "--", "T_PREDEC", TC_UNARY, L_CHARS(2), NULL, NULL
};

static Lex_token token_postinc = {
  T_POSTINC, "++", "T_POSTINC", TC_UNARY, L_CHARS(2), NULL, NULL
};

static Lex_token token_postdec = {
  T_POSTDEC, "--", "T_POSTDEC", TC_UNARY, L_CHARS(2), NULL, NULL
};

static Lex_token token_intval = {
  T_INTEGER, NULL, "T_INTEGER", TC_OPERAND, L_CHARS(0), NULL, NULL
};

static Lex_token token_string = {
  T_STRING, NULL, "T_STRING", TC_OPERAND, L_CHARS(0), NULL, NULL
};

static Lex_token token_realval = {
  T_REAL, NULL, "T_REAL", TC_OPERAND, L_CHARS(0), NULL, NULL
};

static Lex_token token_literal = {
  T_LITERAL, NULL, "T_LITERAL", TC_OPERAND, L_CHARS(0), NULL, NULL
};

static Lex_token token_func = {
  T_FUNC, NULL, "T_FUNC", TC_OPERAND, L_CHARS(0), NULL, NULL
};

static int is_operand(Lex_state *e, Expr_result *res);
static int heir(Lex_state *e, Expr_result *result);
static int heir2(Lex_state *e, Expr_result *result);
static int heir3(Lex_state *e, Expr_result *result);
static int heir4(Lex_state *e, Expr_result *result);
static int heir5(Lex_state *e, Expr_result *result);
static int heir6(Lex_state *e, Expr_result *result);
static int heir7(Lex_state *e, Expr_result *result);
static int heir8(Lex_state *e, Expr_result *result);
static int heir9(Lex_state *e, Expr_result *result);
static int heir10a(Lex_state *e, Expr_result *result);
static int heir10b(Lex_state *e, Expr_result *result);
static int heir11(Lex_state *e, Expr_result *result);
static int heir12(Lex_state *e, Expr_result *result);
static int heir13a(Lex_state *e, Expr_result *result);
static int heir13b(Lex_state *e, Expr_result *result);
static int heir15(Lex_state *e, Expr_result *result);
static int heir17(Lex_state *e, Expr_result *result);
static int eval1(Lex_state *e, Expr_result *res, Token op, Expr_result *r1);
static int eval2(Lex_state *e, Expr_result *res, Token op, Expr_result *r1,
				 Expr_result *r2);
#ifdef NOTDEF
static int eval3(Lex_state *e, Expr_result *res, Token op, Expr_result *r1,
				 Expr_result *r2, Expr_result *r3);
#endif
static int do_if(Lex_state *e, Expr_result *result);
static int eval_function(char *name, Lex_state *e, Arglist *args,
						 Expr_result *result);
static int funargs(Lex_state *e, Arglist **head);
static int indexset(char *s1, char *s2);

static int func_ack(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_contains_any(Lex_state *e, Arglist *arglist,
							 Expr_result *result);
static int func_dacs_admin(Lex_state *e, Arglist *arglist,
						   Expr_result *result);
static int func_mime_decode(Lex_state *e, Arglist *arglist,
							Expr_result *result);
static int func_mime_encode(Lex_state *e, Arglist *arglist,
							Expr_result *result);
static int func_eval(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_exec(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_exit(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_file(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_file_group(Lex_state *e, Arglist *arglist,
						   Expr_result *result);
static int func_file_owner(Lex_state *e, Arglist *arglist,
						   Expr_result *result);
static int func_from(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_get(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_http(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_index(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_info(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_ldap(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_length(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_pathname(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_print(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_printf(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_redirect(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_regmatch(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_regsub(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_setvar(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_strftime(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_strtr(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_subset(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_substr(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_system(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_time(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_url_decode(Lex_state *e, Arglist *arglist,
						   Expr_result *result);
static int func_url_encode(Lex_state *e, Arglist *arglist,
						   Expr_result *result);
static int func_user(Lex_state *e, Arglist *arglist, Expr_result *result);
static int func_user_revoke(Lex_state *e, Arglist *arglist,
							Expr_result *result);
static int func_vfs(Lex_state *e, Arglist *arglist, Expr_result *result);

/*
 * See validate_function().
 * The first field is the name by which the function is invoked.
 * The second field is the arity; -1 means it's variadic.
 * The third field is the argument descriptor string; the nth character
 * classifies the nth argument:
 * '-': leave the argument as-is
 * 's': convert the argument to a T_STRING if necessary
 * 'S': the argument must already be a T_STRING
 * 'i': convert the argument to a T_INTEGER if necessary
 * 'I': the argument must already be a T_INTEGER
 * 'r': convert the argument to a T_REAL if necessary
 * 'R': the argument must already be a T_REAL
 * '*': means zero or more arguments may follow but quit validating;
 *      if present, this must be last
 * If the arity is -1, the nth element of the descriptor string is meaningful
 * only if there really are that many arguments; e.g., "S" means that if
 * there is one argument and implies there can be no arguments or exactly 1
 * argument.
 * An empty descriptor means leave all arguments as-is.
 *
 * XXX We might also support dynamically loaded functions by providing
 * a way for the administrator to give the callable name, its arity and
 * arg spec, and its library name and the file containing the function.
 * We'd then use dlopen() etc. to link in the function.
 */
Func functions[] = {
  { "ack",          -1,  "S*",     func_ack },
  { "contains_any", -1,  "SSS*",  func_contains_any },
  { "dacs_admin",    0,  "",      func_dacs_admin },
  { "eval",          1,  "S",     func_eval },
  { "exec",         -1,  "*",     func_exec },
  { "exit",          1,  "-",     func_exit },
  { "file",         -1,  "S*",    func_file },
  { "file_group",   -1,  "S",     func_file_group },
  { "file_owner",   -1,  "S",     func_file_owner },
  { "from",          1,  "s",     func_from },
  { "get",          -1,  "S*",    func_get },
  { "http",         -1,  "SS*",   func_http },
  { "index",         2,  "SS",    func_index },
  { "info",         -1,  "S*",    func_info },
  { "ldap",         -1,  "ss*",   func_ldap },
  { "length",        1,  "s",     func_length },
  { "mime_decode",   1,  "S",     func_mime_decode },
  { "mime_encode",   1,  "S",     func_mime_encode },
  { "pathname",      3,  "SSs*",  func_pathname },
  { "print",        -1,  "S*",    func_print },
  { "printf",       -1,  "S*",    func_printf },
  { "redirect",      1,  "S",     func_redirect },
  { "regmatch",     -1,  "SS*",   func_regmatch },
  { "regsub",       -1,  "SSS*",  func_regsub },
  { "return",        1,  "-",     func_exit },
  { "revoke",        1,  "S",     func_user_revoke },
  { "setvar",       -1,  "SS*",   func_setvar },
  { "vfs",          -1,  "ss*",   func_vfs },
  { "strftime",      1,  "S",     func_strftime },
  { "strtr",        -1,  "SS*",   func_strtr },
  { "subset",       -1,  "SSS*",  func_subset },
  { "substr",        3,  "SII",   func_substr },
  { "system",        1,  "S",     func_system },
  { "time",         -1,  "s*",    func_time },
  { "url_decode",    1,  "S",     func_url_decode },
  { "url_encode",    1,  "S",     func_url_encode },
  { "user",          1,  "S",     func_user },
  { NULL,            0,  NULL,    NULL }
};

static int do_lexdump = 0;

static Lex_token *
token_lookahead(Lex_state *e, int n)
{
  Lex_token *t;

  t = dsvec_ptr(e->lexptr, e->ctn + n, Lex_token *);

  return(t);
}

static int
token_matches(Lex_state *e, int n, Token token)
{
  Lex_token *t;

  t = dsvec_ptr(e->lexptr, e->ctn + n, Lex_token *);
  if (t != NULL && t->token == token)
	return(1);

  return(0);
}

static Lex_token *
token_peek(Lex_state *e, int n)
{
  int ind;
  Lex_token *t;

  ind = e->ctn + n;
  if (ind < 0 || ind >= dsvec_len(e->lexptr))
	t = NULL;
  else
	t = dsvec_ptr(e->lexptr, ind, Lex_token *);

  return(t);
}

static void
seterr(Lex_state *e, Expr_result *result, Acs_expr_result err, char *msg)
{
  Lex_token *t;
  Ds ds;

  e->do_eval = 0;
  result->err = err;
  if (result->errmsg == NULL) {
	ds_init(&ds);
	ds_asprintf(&ds, "%s", msg);

	t = token_lookahead(e, 0);
	if (t->token == T_EOI)
	  ds_asprintf(&ds, ", token=end-of-input");
	else if (t->token == T_EOS)
	  ds_asprintf(&ds, ", token=end-of-statement");
	else if (t->lexeme != NULL)
	  ds_asprintf(&ds, ", token=\"%s\"", t->lexeme);
	else
	  ds_asprintf(&ds, ", token=%d", t->token);

	if (t->startaddr != NULL)
	  ds_asprintf(&ds, " at offset=%d", t->startaddr - e->expr);

	if ((t = token_peek(e, -1)) != NULL)
	  ds_asprintf(&ds, " prev_token=\"%s\"", t->lexeme);
	else
	  ds_asprintf(&ds, " prev_token=start-of-input");

	result->errmsg = ds_buf(&ds);
  }
}

static void
seterr_e(Lex_state *e, Expr_result *result, char *msg)
{

  seterr(e, result, ACS_EXPR_EVAL_ERROR, msg);
}

#ifdef NOTDEF
static void
seterr_l(Lex_state *e, Expr_result *result, char *msg)
{

  seterr(e, result, ACS_EXPR_LEXICAL_ERROR, msg);
}
#endif

static void
seterr_s(Lex_state *e, Expr_result *result, char *msg)
{

  seterr(e, result, ACS_EXPR_SYNTAX_ERROR, msg);
}

typedef enum Comment_style {
  COMMENT_NONE    = 0,
  COMMENT_SLASHES = 1,
  COMMENT_C       = 2
} Comment_style;

/*
 * Given an expression, remove all commented out areas.
 * If there isn't at least one comment, return NULL (so that the original
 * copy of the expression can be used), otherwise return a comment-free
 * copy of the expression.
 *
 * We recognize C++ style comments ("// a comment until end-of-line").
 */
Ds *
acs_elide_comments(char *expr)
{
  int quote_ch;
  char *comment_start, *p, *start_p;
  Ds *ds;
  Comment_style cstyle;

  if (expr == NULL)
	return(NULL);

  start_p = p = expr;
  comment_start = NULL;
  cstyle = COMMENT_NONE;

  /* Watch for the comment initiator characters strings. */
  quote_ch = '\0';
  while (*p != '\0') {
	if (*p == '"' || *p == '\'') {
	  if (*p == quote_ch)
		quote_ch = '\0';
	  else
		quote_ch = *p;
	}
	else if (quote_ch == '\0'
			 && ((*p == '/' && *(p + 1) == '/')
				 || (*p == '/' && *(p + 1) == '*'))) {
	  comment_start = p;
	  if (*(p + 1) == '/')
		cstyle = COMMENT_SLASHES;
	  else
		cstyle = COMMENT_C;
	  break;
	}
	p++;
  }

  if (comment_start == NULL)
	return(NULL);

  /*
   * There is at least one comment.
   * Copy whatever precedes the start of comment.
   */
  ds = ds_init(NULL);
  if (comment_start - start_p) {
	ds_concatn(ds, start_p, comment_start - start_p);
	ds_concatc(ds, '\n');
  }

  while (*p != '\0') {
	/* We're at the start of a comment. */
	p += 2;

	/* Skip (elide) until the end of comment (end of line or buffer. */
	while (*p != '\0') {
	  if ((cstyle == COMMENT_SLASHES && *p == '\n')
		  || (cstyle == COMMENT_C && *p == '*' && *(p + 1) == '/'))
		break;
	  p++;
	}

	if (*p == '\0')
	  break;

	if (cstyle == COMMENT_SLASHES)
	  p++;
	else
	  p += 2;
	cstyle = COMMENT_NONE;

	/* Copy until the start of the next comment or end of buffer. */
	quote_ch = '\0';
	while (*p != '\0') {
	  if (*p == '"' || *p == '\'') {
		if (*p == quote_ch)
		  quote_ch = '\0';
		else
		  quote_ch = *p;
	  }
	  else if (quote_ch == '\0' &&
			   ((*p == '/' && *(p + 1) == '/')
				|| (*p == '/' && *(p + 1) == '*'))) {
		if (*(p + 1) == '/')
		  cstyle = COMMENT_SLASHES;
		else
		  cstyle = COMMENT_C;
		break;
	  }
	  ds_concatc(ds, (int) *p);
	  p++;
	}
  }

  if (cstyle != COMMENT_NONE)
	log_msg((LOG_ERROR_LEVEL, "Unterminated comment found"));

  return(ds);
}

static void
lexdump(char *str, Dsvec lexptr, int n)
{
  int i;
  Lex_token *t;

  for (i = 0; i < n; i++) {
	if ((t = dsvec_ptr(&lexptr, i, Lex_token *)) == NULL)
	  break;
	if (t->lexeme != NULL)
	  fprintf(stderr, "token %2d: \"%s\"", i, strquote(t->lexeme, "\""));
	else
	  fprintf(stderr, "token %2d: %d", i, t->token);

	if (t->token_name != NULL)
	  fprintf(stderr, " (%s)", t->token_name);
	if (t->startaddr != NULL)
	  fprintf(stderr, " at offset %u", (unsigned int) (t->startaddr - str));
	fprintf(stderr, "\n");
  }
}

Lex_token *
lookup_token(Token t)
{
  int i;

  for (i = 0; tokens[i].token != T_UNDEF; i++) {
	if (tokens[i].token == t)
	  return(&tokens[i]);
  }

  if (token_eoi.token == t)
	return(&token_eoi);
  if (token_unary_minus.token == t)
	return(&token_unary_minus);
  if (token_unary_plus.token == t)
	return(&token_unary_plus);
  if (token_preinc.token == t)
	return(&token_preinc);
  if (token_predec.token == t)
	return(&token_predec);
  if (token_postinc.token == t)
	return(&token_postinc);
  if (token_postdec.token == t)
	return(&token_postdec);
  if (token_string.token == t)
	return(&token_string);
  if (token_intval.token == t)
	return(&token_intval);
  if (token_realval.token == t)
	return(&token_realval);
  if (token_literal.token == t)
	return(&token_literal);
  if (token_func.token == t)
	return(&token_func);

  return(NULL);
}
	
/*
 * An intval is a string of decimal digits (no sign).
 * If there's at least one decimal digit, set *VAL to the value of
 * the integer and ENDP to point to the first non-digit character, and
 * return 1.
 * strtol(3) recognizes octal, decimal, and hex values.
 * Return 0 if no integer is found.
 * Return -1 if the integer is invalid.
 */
int
is_intval(char *str, long *val, char **endp)
{
  long v;

  if (!isdigit((int) *str))
	return(0);

  /* A base of 0 means to expect a C-style integer constant. */
  v = strtol(str, endp, 0);
  if ((v == LONG_MAX || v == LONG_MIN) && errno == ERANGE)
    return(-1);

  if (val != NULL)
	*val = v;

  return(1);
}

static int
intval(char *str, Expr_result *res)
{
  long val;

  if (is_intval(str, &val, NULL) != 1)
    return(-1);

  res->value.token = T_INTEGER;
  res->value.val.intval = val;
  res->err = 0;
  res->errmsg = NULL;
  return(0);
}

/*
 * A realval is an optional string of decimal digits (possibly signed)
 * followed by a period and at least one digit or an 'e' or 'E' followed
 * by at least one digit.
 * If the legal syntax appears, set *VAL to the value of the real number and
 * ENDP to point to the first character not part of the number, and
 * return 1.
 * Return 0 if no real is found.
 * Return -1 if the real is invalid.
 */
static int
is_realval(char *str, double *val, char **endp)
{
  char *end;
  double v;

  if (!isdigit((int) *str) && *str != '.')
    return(0);

  v = strtod(str, &end);
  if (end == str)
    return(0);
  if ((v == HUGE_VAL || v == -HUGE_VAL || v == 0.0) && errno == ERANGE)
    return(-1);

  if (endp != NULL)
	*endp = end;
  *val = v;
  return(1);
}

static int
is_exponent(char *str)
{
  char *p;

  p = str;
  if (*p != 'e' && *p != 'E')
    return(0);
  p++;

  if (*p == '+' || *p == '-')
    p++;

  if (isdigit((int) *p))
    return(1);

  return(0);
}

static Value *
init_value(Value *v, Token token, ...)
{
  va_list ap;
  Value *value;

  if (v != NULL)
	value = v;
  else
	value = ALLOC(Value);

  value->token = token;
  value->is_var_reference = 0;
  value->is_quoted = 0;
  value->varname = NULL;

  va_start(ap, token);
  if (token == T_REAL)
	value->val.realval = va_arg(ap, double);
  else if (token == T_INTEGER)
	value->val.intval = va_arg(ap, long);
  else if (token == T_STRING || token == T_FUNC || token == T_LITERAL)
	value->val.strval = va_arg(ap, char *);
  else if (token == T_UNDEF)
	value->val.strval = NULL;

  va_end(ap);

  return(value);
}

static Value *
copy_value(Value *new_value, Value *old_value)
{
  Value *value;

  if (new_value != NULL)
	value = new_value;
  else
	value = ALLOC(Value);

  value->token = old_value->token;
  value->is_var_reference = old_value->is_var_reference;
  value->varname = (old_value->varname == NULL)
	? NULL : strdup(old_value->varname);
  value->is_quoted = old_value->is_quoted;
  if (old_value->token == T_REAL)
	value->val.realval = old_value->val.realval;
  else if (old_value->token == T_INTEGER)
	value->val.intval = old_value->val.intval;
  else if (old_value->token == T_STRING)
	value->val.strval = (old_value->val.strval == NULL)
	  ? NULL : strdup(old_value->val.strval);

  return(value);
}

static Expr_result *
init_expr_result(Expr_result *result)
{
  Expr_result *er;

  if (result == NULL)
	er = ALLOC(Expr_result);
  else
	er = result;
  init_value(&er->value, T_UNDEF);
  er->err = ACS_EXPR_FALSE;
  er->errmsg = NULL;

  return(er);
}

static int
value_init(Expr_result *new_res, Expr_result *old_res)
{

  init_expr_result(new_res);
  new_res->err = old_res->err;
  new_res->errmsg = old_res->errmsg;

  if (new_res->err)
	return(-1);
  return(0);
}

/*
 * If the initial part of STR is an integer or real number, create a token,
 * set ENDP to the first character not part of the number, and return 1.
 * ENDP may be NULL if no end pointer is required.
 * TP may be NULL if no token is required.
 * If no number can be constructed, return 0; if a number appears but is
 * invalid (e.g., too big), return -1; otherwise, return 1.
 */
static int
is_number(char *str, Lex_token **tp, char **endp)
{
  int st;
  char *p;
  long intval;
  double realval;
  Lex_token *t;

  p = str;
  if (*p == '.') {
	if (!isdigit((int) *(p + 1)))
	  return(0);
  }
  else {
	while (isdigit((int) *p))
	  p++;
	if (p == str)
	  return(0);
  }

  /* This is a number, but is it an integer or a real? */
  if (*p == '.' || is_exponent(p)) {
    if ((st = is_realval(str, &realval, endp)) == -1)
      return(-1);
    if (st == 1) {
	  if (tp == NULL)
		return(1);
      t = ALLOC(Lex_token);
      t->token = T_REAL;
	  t->token_name = "T_REAL";
      t->tc = TC_OPERAND;
	  t->startaddr = str;
      t->value = init_value(NULL, T_REAL, realval);
      *tp = t;
      return(1);
    }
    /* Fall through and guess that it's supposed to be an integer */
  }

  if ((st = is_intval(str, &intval, endp)) == 1) {
	if (tp == NULL)
	  return(1);
	t = ALLOC(Lex_token);
	t->token = T_INTEGER;
	t->token_name = "T_INTEGER";
	t->tc = TC_OPERAND;
	t->startaddr = str;
	t->value = init_value(NULL, T_INTEGER, intval);
	*tp = t;
	return(1);
  }

  return(st);
}

static int
force_boolean(Expr_result *r, int *boolp)
{
  int b;

  b = 0;
  if (r->value.token == T_STRING) {
	if (r->value.val.strval[0] != '\0')
	  b = 1;
  }
  else if (r->value.token == T_INTEGER) {
	if (r->value.val.intval != 0L)
	  b = 1;
  }
  else if (r->value.token == T_REAL) {
	if (r->value.val.realval != 0.0)
	  b = 1;
  }
  else
	return(-1);

  *boolp = b;
  return(0);
}

static int
force_integer(Expr_result *r, long *ip)
{
  long i;

  if (r->value.token == T_REAL)
	return(-1);
  else if (r->value.token == T_INTEGER)
	i = r->value.val.intval;
  else if (r->value.token == T_STRING) {
	char *endp;

	if (is_intval(r->value.val.strval, &i, &endp) != 1 || *endp != '\0')
	  return(-1);
  }
  else
	return(-1);

  *ip = i;
  return(0);
}

static int
force_real(Expr_result *r, double *dp)
{
  double d;

  if (r->value.token == T_REAL)
	d = r->value.val.realval;
  else if (r->value.token == T_INTEGER)
	d = (double) r->value.val.intval;
  else if (r->value.token == T_STRING) {
	char *endp;

	if (is_realval(r->value.val.strval, &d, &endp) != 1 || *endp != '\0')
	  return(-1);
  }
  else
	return(-1);

  *dp = d;
  return(0);
}

static int
force_string(Expr_result *r, char **strp)
{
  char *str;

  if (r->value.token == T_STRING || r->value.token == T_LITERAL)
	str = r->value.val.strval;
  else if (r->value.token == T_REAL)
	str = ds_xprintf("%f", r->value.val.realval);
  else if (r->value.token == T_INTEGER)
	str = ds_xprintf("%ld", r->value.val.intval);
  else
	return(-1);

  *strp = str;
  return(0);
}

#ifdef NOTDEF
static int
is_real_operand(Expr_result *r, double *rvalp)
{
  char *endp;

  if (r->value.token == T_REAL) {
	*rvalp = r->value.val.realval;
	return(1);
  }
  if (r->value.token == T_STRING
	  && is_realval(r->value.val.strval, rvalp, &endp) == 1 && *endp == '\0')
	return(1);

  return(0);
}

static int
is_integer_operand(Expr_result *r, long *ip)
{

  if (r->value.token == T_INTEGER) {
	*ip = r->value.val.intval;
	return(1);
  }
  if (r->value.token == T_STRING) {
	char *endp;

	if (is_intval(r->value.val.strval, ip, &endp) == 1 && *endp == '\0')
	  return(1);
  }

  return(0);
}
#endif

static int
is_valid_bareword(char *str, char **namep, char **nextp)
{
  char *p;

  p = str;
  if (!isalpha((int) *p))
    return(0);
  p++;
  while (isalnum((int) *p) || *p == '_')
    p++;

  if (nextp != NULL)
	*nextp = p;

  if (namep != NULL)
	*namep = strndup(str, p - str);

  return(1);
}

static int
is_valid_function_name(char *str, char **namep, char **nextp)
{
  char *endp, *p;

  p = str;
  if (!isalpha((int) *p) && *p != '_')
	return(0);
  p++;

  while (isalnum((int) *p) || *p == '_')
    p++;

  if (*p == ':' && *(p + 1) == ':') {
	if (*(p + 2) == '\0')
	  return(0);
	p += 2;
  }

  while (isalnum((int) *p) || *p == '_')
    p++;

  endp = NULL;
  if (*p == ' ' || *p == '\t')
	endp = p++;

  while (*p == ' ' || *p == '\t')
	p++;

  if (*p != '(')
	return(0);
  if (endp == NULL)
	endp = p;

  if (namep != NULL)
	*namep = strndup(str, endp - str);
	
  if (nextp != NULL)
	*nextp = p;

  return(1);
}

/*
 * Test if STR looks like a function name or another literal symbol.
 * If it does, create and return a token, set ENDP to point to the first
 * character past the name, and return 1.
 * Otherwise, return 0.
 */
static int
is_literal(char *str, Lex_token **tp, char **endp)
{
  char *bname, *fname, *p;
  Lex_token *t;

  if (is_valid_function_name(str, &fname, &p)) {
	t = ALLOC(Lex_token);
	*t = token_func;
	t->startaddr = str;
	t->value = init_value(NULL, T_FUNC, fname);
  }
  else if (is_valid_bareword(str, &bname, &p)) {
	/* This is similar to what perl calls a "bareword". */
	t = ALLOC(Lex_token);
	*t = token_literal;	
	t->startaddr = str;
	t->value = init_value(NULL, T_LITERAL, bname);
  }
  else
	return(0);

  *tp = t;
  *endp = p;

  return(1);
}

/*
 * Lexeme scanner
 * Starting at STR, examine characters to determine the next input token.
 * PREV_T points to the previous token (if NULL, this is the first token).
 * Set ENDP to point to the character following a valid lexeme.
 * Return a pointer to the token or NULL on error.
 */
static Lex_token *
lexeme_scan(char *str, Lex_token *prev_t, char **endp)
{
  int len, quote_ch, st;
  Lex_token *newt, *t;

  for (t = &tokens[0]; t->token != T_UNDEF; t++) {
	len = lexeme_length(t);
	if ((IS_L_CHARS(t->type_length) && strneq(str, t->lexeme, len))
		|| (IS_L_ALPHA(t->type_length) && strneq(str, t->lexeme, len)
			&& !isalpha((int) str[len]))) {
	  /*
	   * Special handling for the binary period operator.
	   */
	  if (prev_t == NULL || prev_t->token == T_LPAREN || is_operator(prev_t)) {
		if (t->token == T_DOT)
		  continue;
	  }
	  *endp = str + len;
	  newt = ALLOC(Lex_token);
	  *newt = *t;
	  newt->startaddr = str;
	  return(newt);
	}
  }

  /* It must be a number or a string. */

  if ((st = is_number(str, &t, endp)) == -1)
	return(NULL);
  else if (st == 1)
	return(t);

  /*
   * XXX We might add backquotes/backticks (command interpolation) here
   * as a new type of string.  We'd need to do variable substitution within
   * the backquoted region and but not any other kind of parsing.  Redirection
   * and such would be supported.  The output
   * of the command, if any, would be interpolated.
   * The syntax would be similar to perl's rather than the shell's,
   * so command interpolation would not be embedded within other strings.
   * Note that this feature has security implications, so certain
   * restrictions might be enforced.  Also, might the command be allowed to
   * be a built-in?
   */
  if (*str == '"' || *str == '\'') {
	char *p;
	Ds *s;

	quote_ch = *str;
	s = ds_init(NULL);
	p = str + 1;
	while (1) {
	  if (*p == '\\') {
#ifdef NOTDEF
		/* Handle C-style character constants. */
		p++;
		if (*p == '\0')	{ /* Ends with a backslash? */
		  return(NULL);
		}

		if (is_octdigit((int) *p)) {
		  cc = 0;
		  for (i = 0; i < 3; i++) {
			if (!is_octdigit((int) *p))
			  break;
			cc = cc * 8 + octdigit_val((int) *p);
			p++;
		  }
		  ds_appendc(s, cc);
		}
		else {
		  /* Harbison & Steele 2.7.6 */
		  switch ((int) *p) {
		  case 'a':
			cc = '\a'; break;
		  case 'b':
			cc = '\b'; break;
		  case 'f':
			cc = '\f'; break;
		  case 'n':
			cc = '\n'; break;
		  case 'r':
			cc = '\r'; break;
		  case 't':
			cc = '\t'; break;
		  case 'v':
			cc = '\v'; break;
		  default:
			cc = *p; break;
		  }

		  ds_appendc(s, cc);
		  p++;
		}
#else
		if (*(p + 1) == '\0')	{ /* Ends with a backslash? */
		  return(NULL);
		}
		if (*(p + 1) == quote_ch || *(p + 1) == '\\') {
		  p++;
		  ds_appendc(s, (int) *p++);
		}
		else {
		  ds_appendc(s, (int) *p++);
		  ds_appendc(s, (int) *p++);
		}
#endif
	  }

	  else if (*p == quote_ch) {
		p++;
		ds_appendc(s, '\0');
		break;
	  }
	  else if (*p == '\0')
		return(NULL);
	  else {
		ds_appendc(s, *p);
		p++;
	  }
	}
	newt = ALLOC(Lex_token);
	*newt = token_string;
	newt->startaddr = str;
    newt->value = init_value(NULL, T_STRING, ds_buf(s));
	newt->value->is_quoted = (quote_ch == '\'');
	*endp = p;
	return(newt);
  }

  if (*str == '$' && *(str + 1) == '{') {
	char *p;
	Ds ds;

	/* A simple variable reference */
	ds_init(&ds);
	ds_appendc(&ds, '$');
	ds_appendc(&ds, '{');
	p = str + 2;
	while (*p != '}' && *p != '\0') {
	  if (*p == '\\' && *(p + 1) == '}')		
		ds_appendc(&ds, *p++);
	  ds_appendc(&ds, *p++);
	}
	if (*p == '\0')
	  return(NULL);
	ds_appendc(&ds, '}');
	ds_appendc(&ds, '\0');

	newt = ALLOC(Lex_token);
	*newt = token_string;
	newt->startaddr = str;
    newt->value = init_value(NULL, T_STRING, ds_buf(&ds));
	newt->value->is_var_reference = 1;
	newt->value->varname = ds_buf(&ds);
	*endp = p + 1;
	return(newt);
  }

  if (is_literal(str, &t, endp))
	return(t);

  return(NULL);
}

/*
 * Lexical analyzer
 * Return the number of tokens if ok, -1 otherwise.
 */
static int
lexan(Lex_state *e, Expr_result *result)
{
  char *input, *p, *nextp;
  Lex_token *t, *t_prev;
  Acs_environment *env;
  Dsvec *lexptr;
  static char *c[] = { "unary", "binary", "ternary", "operand", "spec" };

  input = e->expr;
  env = e->env;
  lexptr = e->lexptr;

  p = input;
  t_prev = NULL;

  while (*p != '\0') {
	/* Skip any leading white space */
	if (*p == ' ' || *p == '\t' || *p == '\n') {
	  p++;
	  continue;
	}

	if ((t = lexeme_scan(p, t_prev, &nextp)) == NULL) {
	  result->err = ACS_EXPR_LEXICAL_ERROR;
	  result->errmsg = ds_xprintf("Character 0%o at offset %d",
								  *p, p - input);
	  return(-1);
	}
    t->lexeme = (char *) malloc(nextp - p + 1);
    strcpyn(t->lexeme, p, nextp - p);
	dsvec_add_ptr(lexptr, (void *) t);

	log_msg((LOG_TRACE_LEVEL, "lexeme: %s\"%s\" (t=%d, c=%s)",
			 (t->value != NULL && t->value->is_var_reference) ? "var " : "",
			 t->lexeme, t->token, c[t->tc]));

	t_prev = t;
	p = nextp;
  }

  dsvec_add_ptr(lexptr, (void *) &token_eoi);

  return(dsvec_len(lexptr));
}

char *
acs_format_result(Expr_result *res)
{
  char *s;

  if (force_string(res, &s) == -1)
	return(NULL);

  return(s);
}

static int
is_true_value(Value *v)
{

  if (v->token == T_INTEGER || v->token == T_STRING || v->token == T_REAL) {
	if ((v->token == T_INTEGER && v->val.intval == 0L)
		|| (v->token == T_STRING && v->val.strval[0] == '\0')
		|| (v->token == T_REAL && v->val.realval == 0.0))
	  return(0);

	return(1);
  }

  return(-1);
}

static Lex_token *
matches(Lex_state *e, Token token)
{
  Lex_token *t;

  t = dsvec_ptr(e->lexptr, e->ctn, Lex_token *);
  if (t->token == T_EOI)
	return(NULL);

  if (t->token == token) {
	e->ctn++;
	return(t);
  }

  return(NULL);
}

/*
 * Process a sequence of zero or more statements.
 * If the first token is a left brace, we expect to stop at the matching
 * right brace; otherwise the block is implicit and we stop at the end-of-input
 * token.
 * The value of the block is that of the last statement, or the
 * empty string if the block is empty.
 */
static int
do_block(Lex_state *e, Expr_result *result)
{
  int need_rbrace, saw_eos;

  if (matches(e, T_LBRACE))
	need_rbrace = 1;
  else
	need_rbrace = 0;

  /* The block looks like "{ }" (it's empty). */
  if (need_rbrace && matches(e, T_RBRACE)) {
	if (e->do_eval)
	  init_value(&result->value, T_STRING, "");

	return(0);
  }

  /* Evaluate the statements in the (non-empty) block sequentially. */
  while (1) {
	if (heir(e, result) == -1)
	  return(-1);

	/*
	 * The only time a T_EOS is unnecessary is if there's just one
	 * statement or if the statement is not a brace-delimited block.
	 * There can always be multiple T_EOS tokens.
	 */
	saw_eos = 0;
	while (matches(e, T_EOS))
	  saw_eos = 1;

	if (token_matches(e, 0, T_EOI)) {
	  /* We've hit the end-of-input. */
	  if (need_rbrace) {
		seterr_s(e, result, "Closing brace is missing");
		return(-1);
	  }
	  break;
	}

	/*
	 * If we've hit the closing brace, we're done.
	 * We don't require a T_EOS in this case.
	 */
	if (need_rbrace && matches(e, T_RBRACE)) {
	  e->block_end++;
	  break;
	}

	if (!e->block_end && !saw_eos) {
	  seterr_s(e, result, "Expected ';'");
	  return(-1);
	}
	e->block_end = 0;
  }

  return(0);
}

static int
func_expr(char *expr, Acs_environment *env, Expr_result *result)
{
  int n;
  Dsvec lexptr;
  Lex_token *t;
  Lex_state e;

  dsvec_init(&lexptr, sizeof(Lex_token *));

  e.env = env;
  e.lexptr = &lexptr;
  e.ctn = 0;
  e.expr = expr;
  e.do_eval = env->do_eval;
  e.block_end = 0;

  if ((n = lexan(&e, result)) == -1)
	return(-1);

  if (do_lexdump || env->trace_level > 1)
	lexdump(expr, lexptr, n);

  /*
   * If there is nothing remaining after lexical analysis, there is isn't
   * really an expression.  Return the result as True because this simplifies
   * processing an empty Allow or Deny clause in an ACL since an empty clause
   * in those contexts is True.
   * An "empty" clause may actually consist of white space or comments.
   */
  if (n == 0 || (n == 1 && token_matches(&e, 0, T_EOI))) {
	result->value.token = T_INTEGER;
	result->value.val.intval = 1;
	result->err = ACS_EXPR_TRUE;
	result->errmsg = NULL;

	return(0);
  }

  init_value(&result->value, T_UNDEF);
  result->err = ACS_EXPR_FALSE;
  result->errmsg = NULL;

  /* A block is disallowed at the top level. */
  if (token_matches(&e, 0, T_LBRACE)) {
	t = dsvec_ptr(&lexptr, e.ctn, Lex_token *);
	result->err = ACS_EXPR_SYNTAX_ERROR;
	result->errmsg = ds_xprintf("func_expr: syntax error, token = '%s' (%d)",
								t->lexeme == NULL ? "T_EOI" : t->lexeme,
								t->token);
	return(-1);
  }

  /* Do a syntax checking pass. */
  e.do_eval = 0;
  if (do_block(&e, result) == -1) {
	t = dsvec_ptr(&lexptr, e.ctn, Lex_token *);
	if (result->errmsg == NULL) {
	  result->err = ACS_EXPR_SYNTAX_ERROR;
	  result->errmsg = ds_xprintf("func_expr: syntax error, token = '%s' (%d)",
								  t->lexeme == NULL ? "T_EOI" : t->lexeme,
								  t->token);
	}
	return(-1);
  }
  log_msg((LOG_TRACE_LEVEL, "Syntax check succeeded"));

  if (env->do_eval) {
	/* Evaluate the expression(s), if required. */
	e.do_eval = 1;
	e.ctn = 0;
	e.block_end = 0;
	init_value(&result->value, T_UNDEF);
	result->err = ACS_EXPR_FALSE;
	result->errmsg = NULL;
	if (do_block(&e, result) == -1) {
	  t = dsvec_ptr(&lexptr, e.ctn, Lex_token *);
	  if (result->errmsg == NULL) {
		result->err = ACS_EXPR_EVAL_ERROR;
		result->errmsg = ds_xprintf("func_expr: eval error, token = '%s' (%d)",
									t->lexeme == NULL ? "T_EOI" : t->lexeme,
									t->token);
	  }
	  return(-1);
	}
  }

  return(0);
}

static char *reserved_namespaces[] = {
  "DACS", "Args", "Env", NULL
};

int
is_reserved_namespace(char *ns)
{
  int i;

  if (!is_valid_varspace_name(ns))
	return(1);

#ifdef DEFAULT_NAMESPACE
  if (ns == NULL)		/* It must be the default namespace */
	ns = DEFAULT_NAMESPACE;
#endif

  for (i = 0; reserved_namespaces[i] != NULL; i++) {
	if (streq(ns, reserved_namespaces[i]))
	  return(1);
  }

  if (streq(ns, "Conf") && conf_var_ns_user_readonly)
	return(1);

  return(0);
}

#include <setjmp.h>

static int got_jmp_env = 0;
static jmp_buf jmp_env;

/*
 * This is the primary entry point for evaluating an expression.
 * We assume that the expression has already been preprocessed so that
 * all comments have been removed.
 */
Acs_expr_result
acs_expr(char *expr, Acs_environment *env, Expr_result *result)
{
  int jump_st, st;
  Expr_result *res;

  log_msg((LOG_TRACE_LEVEL, "acs_expr:\n%s", expr));

  res = init_expr_result(result);

  /*
   * This can be called recursively, but we only want to return to the
   * top level.
   */
  jump_st = 0;
  if (!got_jmp_env)
	jump_st = setjmp(jmp_env);

  if (jump_st == 0) {
	got_jmp_env++;
	st = func_expr(expr, env, res);
  }
  else {
	got_jmp_env = 0;
	log_msg((LOG_DEBUG_LEVEL, "longjmp called: setjmp returned %d", jump_st));
	return(ACS_EXPR_TRUE);
  }
  got_jmp_env--;

  if (st == -1) {
	char *err;

  fail:
	if (res->err == ACS_EXPR_SYNTAX_ERROR)
	  err = "Syntax error";
	else if (res->err == ACS_EXPR_LEXICAL_ERROR)
	  err = "Lexical error";
	else if (res->err == ACS_EXPR_EVAL_ERROR)
	  err = "Evaluation error";
	else
	  err = "Unknown error";

	if (res->errmsg != NULL)
	  log_msg((LOG_ERROR_LEVEL, "%s: %s", err, res->errmsg));
	else
	  log_msg((LOG_ERROR_LEVEL, "%s", err));
	return(res->err);
  }

  if (env->do_eval && (st = is_true_value(&res->value)) == -1) {
	res->err = ACS_EXPR_EVAL_ERROR;
	res->errmsg = ds_xprintf("Invalid boolean result value, expr=%s", expr);
	goto fail;
  }

  if (st == 0)
	return(ACS_EXPR_FALSE);

  return(ACS_EXPR_TRUE);
}

Acs_expr_result
acs_expr_string(char *expr, Acs_expr_ns_arg *ns_args, char **result_str)
{
  int i;
  Acs_environment env;
  Acs_expr_result st;
  Expr_result result;

  env.url = NULL;
  env.do_eval = 1;
  env.acl_rule = NULL;
  env.notices = NULL;
  env.redirect = NULL;
  env.namespaces = NULL;
  env.trace_level = trace_level;
  env.credentials = NULL;

  for (i = 0; ns_args[i].name != NULL; i++)
	var_ns_new(&env.namespaces, ns_args[i].name, ns_args[i].kwv);

  log_msg((LOG_DEBUG_LEVEL, "Evaluating: %s", expr));
  st = acs_expr(expr, &env, &result);
  log_msg((LOG_DEBUG_LEVEL, "Evaluation result: %d", st));

  if (result_str != NULL) {
	if (st == ACS_EXPR_FALSE || st == ACS_EXPR_TRUE)
	  *result_str = acs_format_result(&result);
	else
	  *result_str = NULL;
  }

  return(st);
}

/*
 * Given a vector of lines, each one a separate expression, evaluate each
 * one in turn until one evaluates to a non-empty string; return that string.
 * A NULL entry terminates the list.
 * Stop immediately if an evaluation error occurs. 
 */
Acs_expr_result
acs_expr_list_eval(Dsvec *dsv, Acs_expr_ns_arg *ns_args, char **result_str)
{
  int i;
  char *expr, *val;
  Acs_expr_result st;

  for (i = 0; i < dsvec_len(dsv); i++) {
	expr = dsvec_ptr(dsv, i, char *);
	if (expr == NULL)
	  break;
	st = acs_expr_string(expr, ns_args, &val);
	if (st < 0)
	  return(st);
	if (st == ACS_EXPR_TRUE && val != NULL && *val != '\0') {
	  *result_str = val;
	  return(ACS_EXPR_TRUE);
	}
  }

  return(ACS_EXPR_FALSE);
}

/*
 * Read expressions from VFS_URI (or else use ITEM_TYPE), one per line
 * (which may be continued).
 * Then evaluate each one until one evaluates to a non-empty string; set
 * RESULT_STR to its value and return ACS_EXPR_TRUE.
 * If the no such expression is found, return ACS_EXPR_FALSE.
 * If an error occurs, return the error code.
 */
Acs_expr_result
acs_expr_list(char *vfs_uri, char *item_type,
			  Acs_expr_ns_arg *ns_args, char **result_str)
{
  int i;
  char *buffer, *expr;
  Acs_expr_result st;
  Ds line;
  Dsvec *dsv;
  Vfs_handle *h;

  if (vfs_uri != NULL) {
    if ((h = vfs_open_uri(vfs_uri)) == NULL) {
      log_msg((LOG_DEBUG_LEVEL, "Can't open vfs_uri \"%s\"", vfs_uri));
      return(ACS_EXPR_EVAL_ERROR);
    }
  }
  else if (item_type != NULL) {
    if ((h = vfs_open_item_type(item_type)) == NULL) {
      log_msg((LOG_DEBUG_LEVEL, "Can't open item type \"%s\"", item_type));
      return(ACS_EXPR_EVAL_ERROR);
    }
  }
  else
    return(ACS_EXPR_EVAL_ERROR);

  if (vfs_get(h, NULL, (void **) &buffer, NULL) == -1) {
    log_msg((LOG_DEBUG_LEVEL,
             "Could not load expressions from \"%s\" failed",
             item_type != NULL ? item_type : vfs_uri));
	vfs_close(h);
	return(ACS_EXPR_EVAL_ERROR);
  }

  vfs_close(h);

  ds_init(&line);
  line.escnl_flag = 1;
  dsio_set(&line, NULL, buffer, 0, 0);
  dsv = dsvec_load(&line);

  st = acs_expr_list_eval(dsv, ns_args, result_str);

  return(st);
}

static char *
acs_variable_value(Acs_environment *env, char *ns, char *name, int icase)
{
  int ic;
  char *val;
  Kwv *kwv;
  Kwv_pair *v;
  extern int do_conf_resolve;
  extern Kwv *nkwv, *merged_kwv, *xkwv;

  if ((kwv = var_ns_lookup_kwv(env->namespaces, ns)) == NULL)
    return(NULL);

  /*
   * If this is happening during the final stages of configuration
   * directive evaluation, then we need to resolve variable references
   * on the RHS of directives, watching out for recursive definitions.
   * NKWV is the "new kwv" being built from the evaluated RHS of KWV.
   * XKWV is used to keep track of variables that are currently being
   * resolved so that we can detect a loop.
   * XXX This is an ugly kludge that needs to be repaired asap
   */
  if (do_conf_resolve && streq(ns, "Conf")) {
	char *lval, *nval;
	Kwv_pair *pair;
	Acs_expr_result st;
	Expr_result result;

	if (do_conf_resolve < 0)
	  return(NULL);

	/* If we've already resolved this variable, return its value. */
	if ((pair = kwv_lookup(nkwv, name)) != NULL)
	  return(pair->val);

	/*
	 * If the variable is another directive, look up its RHS and evaluate it.
	 * Otherwise check if its a predefined (and pre-evaluated) configuration
	 * variable.
	 */
	if ((lval = kwv_lookup_value(merged_kwv, name)) == NULL) {
	  lval = kwv_lookup_value(kwv, name);
	  return(lval);
	}

	/* Note that NAME is being resolved. */
	if (kwv_add(xkwv, name, "") == NULL) {
	  log_msg((LOG_ERROR_LEVEL,
			   "Config resolution error: recursive evaluation"));
	  do_conf_resolve = -1;
	  return(NULL);
	}

	st = acs_expr(lval, env, &result);
	if (st == ACS_EXPR_LEXICAL_ERROR || st == ACS_EXPR_SYNTAX_ERROR
		|| st == ACS_EXPR_EVAL_ERROR || do_conf_resolve < 0) {
	  log_msg((LOG_ERROR_LEVEL,
			   "Config resolution error: evaluation error"));
	  do_conf_resolve = -1;
	  return(NULL);
	}
	nval = acs_format_result(&result);

	/* Record the variable and its value, delete the note. */
	kwv_add(nkwv, name, nval);
	kwv_delete(xkwv, name);

	return(nval);
  }

  ic = kwv->icase;
  kwv->icase = icase;

  if ((v = kwv_lookup(kwv, name)) != NULL) {
	log_msg((LOG_TRACE_LEVEL, "lookup: %s::%s%s -> \"%s\"",
			 ns == NULL ? "" : ns, name, icase ? " (:i)" : "", v->val));
	val = v->val;
  }
  else {
	log_msg((LOG_TRACE_LEVEL, "lookup failed: %s::%s%s",
			 ns == NULL ? "" : ns, name, icase ? " (:i)" : ""));
	val = NULL;
  }

  if (v != NULL && v->next != NULL)
	log_msg((LOG_WARN_LEVEL, "lookup warning: %s::%s%s has multiple values",
			 ns == NULL ? "" : ns, name, icase ? " (:i)" : ""));

  kwv->icase = ic;
  return(val);
}

/*
 * Lookup callback function for variables.
 * Return the variable's value (a string) or the result of applying the
 * flag(s) to that variable or its value (also a string), or NULL if an
 * error occurs.
 *
 * A namespace is required (the old behaviour was to search for the variable
 * if the namespace was omitted).
 *
 * Flags:
 *   e: if the variable exists, the value is "1" otherwise "0".
 *   i: lookup the variable name ignoring case.
 *   ?: if the variable is not defined or has the empty string as its value,
 *      return the value of the expression that follows the flag
 */
static char *
acs_variable_resolve(Var *var, void *arg)
{
  char *f, *rval, *val;
  Acs_environment *env;
  int test_exists, test_exists_or, ignore_varname_case;
  extern int do_conf_resolve;

  env = (Acs_environment *) arg;
  ignore_varname_case = 0;
  test_exists = 0;
  test_exists_or = 0;

  for (f = var->flags; var->flags != NULL && *f != '\0'; f++) {
	if (*f == 'e')
	  test_exists = 1;
	else if (*f == '?') {
	  test_exists_or = 1;
	  if (test_exists) {
		log_msg((LOG_ERROR_LEVEL, "Invalid variable modifier flags: %s",
				 var->flags));
		return(NULL);
	  }
	  f++;
	  break;
	}
	else if (*f == 'i')
	  ignore_varname_case = 1;
	else {
	  log_msg((LOG_ERROR_LEVEL,
			   "Unrecognized variable modifier flag: %c", *f));
	  return(NULL);
	}
  }

  /* Lookup the variable's value as directed. */
  val = acs_variable_value(env, var->ns, var->varname, ignore_varname_case);

  if (do_conf_resolve == -1)
	return(NULL);

  if (test_exists_or) {
	if (val == NULL || *val == '\0')
	  return(acs_string_operand(f, env));
	return(val);
  }

  if (test_exists) {
	if (val == NULL)
	  rval = strdup("");		/* False */
	else
	  rval = strdup("1");	/* True */
  }
  else {
	if (val == NULL) {
	  if (do_conf_resolve) {
		log_msg((LOG_ERROR_LEVEL,
				 "Config resolution error: undefined variable: %s",
				 var_ns_varname(var->ns, var->varname)));
		do_conf_resolve = -1;
		return(NULL);
	  }
	  rval = strdup("");
	}
	else
	  rval = strdup(val);
  }

  return(rval);
}

/*
 * Evaluate the string STR within context ENV and return the value.
 * Variable references within the string are resolved through the
 * callback function acs_variable_resolve().
 * NULL is returned if an error occurs (such as an error parsing a variable).
 */
char *
acs_string_operand(char *str, Acs_environment *env)
{
  char *expanded_str;

  expanded_str = var_expand(str, acs_variable_resolve, env);
  return(expanded_str);
}

static int
inc_variable(Lex_state *e, char *varname, Token token, Expr_result *result)
{
  int incr;
  char *endp, *newval, *val;
  Expr_result res;
  Kwv *kwv;
  Var *var;

  if ((var = var_parse_name(varname, NULL)) == NULL)
	return(-1);
  else if (var->flags != NULL)
	return(-1);

  if (is_reserved_namespace(var->ns))
	return(-1);

  if (!e->do_eval)
	return(0);

  if ((kwv = var_ns_lookup_kwv(e->env->namespaces, var->ns)) == NULL)
	return(-1);

  if ((val = kwv_lookup_value(kwv, var->varname)) == NULL) {
	log_msg((LOG_ERROR_LEVEL,
			 "Increment/decrement requires initialized variable"));
	return(-1);
  }

  if (is_intval(val, NULL, &endp) == 0 || *endp != '\0'
	  || intval(val, &res) == -1) {
	log_msg((LOG_ERROR_LEVEL, "Increment/decrement requires integer value"));
	return(-1);
  }

  if (token == T_PREDEC || token == T_POSTDEC)
	incr = -1;
  else
	incr = 1;

  if (token == T_PREDEC || token == T_PREINC) {
	res.value.val.intval += incr;
	*result = res;
  }
  else {
	*result = res;
	res.value.val.intval += incr;
  }

  if (force_string(&res, &newval) == -1)
	return(-1);

  kwv_replace(kwv, var->varname, newval);

  return(0);
}

/*
 * Return 1 if the current token is an operand (and evaluate it), 0 otherwise.
 */
static int
is_operand(Lex_state *e, Expr_result *res)
{
  char *val;
  Lex_token *t;

  t = dsvec_ptr(e->lexptr, e->ctn, Lex_token *);
  if (t->token == T_STRING) {
	if (t->value->is_quoted)
	  val = t->value->val.strval;
	else
	  val = acs_string_operand(t->value->val.strval, e->env);
    if (val == NULL)
		return(0);

	if (!e->do_eval)
	  return(1);

	if (t->value->is_var_reference) {
	  res->value.is_var_reference = 1;
	  res->value.varname = strdup(t->value->val.strval);
	}
	else {
	  res->value.is_var_reference = 0;
	  res->value.varname = NULL;
	}

	res->value.token = T_STRING;
	res->value.val.strval = val;
	res->value.is_quoted = t->value->is_quoted;
	return(1);
  }

  if (t->token == T_INTEGER) {
	if (e->do_eval)
	  res->value = *t->value;
	return(1);
  }

  if (t->token == T_REAL) {
	if (e->do_eval)
	  res->value = *t->value;
	return(1);
  }

  if (t->token == T_LITERAL) {
	if (e->do_eval) {
	  res->value.token = T_LITERAL;
	  res->value = *t->value;
	}
	return(1);
  }

  return(0);
}

typedef struct AssignToken {
  Token assign_op;
  Token op;
} AssignToken;

static Token
match_assign_token(Token t)
{
  int i;
  static AssignToken assign[] = {
	{ T_ASSIGN,    T_ASSIGN },
	{ T_PLUS_EQ,   T_PLUS },
	{ T_MINUS_EQ,  T_MINUS },
	{ T_TIMES_EQ,  T_TIMES },
	{ T_DIV_EQ,    T_DIV },
	{ T_MOD_EQ,    T_MOD },
	{ T_BITSHR_EQ, T_BITSHR },
	{ T_BITSHL_EQ, T_BITSHL },
	{ T_BITAND_EQ, T_BITAND },
	{ T_BITXOR_EQ, T_BITXOR },
	{ T_BITOR_EQ,  T_BITOR },
	{ T_DOT_EQ,    T_DOT },
	{ T_UNDEF,     T_UNDEF }
  };

  for (i = 0; assign[i].assign_op != T_UNDEF; i++) {
	if (t == assign[i].assign_op)
	  return(assign[i].op);
  }

  return(T_UNDEF);
}

/*
 * Recursive descent parser
 *
 * Grammar:
 *
 * S  -> EXP | EXP ";" | EXP ";" S
 * EXP -> E
 *
 * E    -> E2   | E2 "," E
 * E2   -> E3   | VAR ASSIGN_OP E2 | IF_ELSEIF_ELSE
 * E3   -> E4   | E4 "?" E ":" E
 * E4   -> E5   | E5 OR E5
 * E5   -> E6   | E6 AND E5
 * E6   -> E7   | E7 "|" E7
 * E7   -> E8   | E8 "^" E8
 * E8   -> E9   | E9 "&" E9
 * E9   -> E10a | E10a EQ_OP E10a
 * E10a -> E10b | E10b REL_OP E10b
 * E10b -> E11  | E11 "." E11
 * E11  -> E12  | E12 "<<" E12 | E12 ">>" E12
 * E12  -> E13a | E13a "+" E13a  | E13a "-" E13a
 * E13a -> E13b | E13b "*" E13b  | E13b "/" E13b | E13b "%" E13b
 * E13b -> E15  | E15 "**" E13a
 * E15  -> E17  | NOT E15 | "~" E15 | "++" VAR | "--" VAR | "+" E | "-" E
 * E17  -> "(" E ")" | VAR "++" | VAR "--" | FUNCTION_CALL | PRIMARY
 *
 * ASSIGN_OP -> "=" | "+=" | "-=" | "*=" | "/=" | "%=" | ">>="
 *   | "<<=" | "&=" | "^=" | "|=" | ".="
 *
 * PRIMARY -> number | string
 * OR      -> "||" | "or"
 * AND     -> "&&" | "and"
 * NOT     -> "!" | "not"
 * EQ_OP   -> "==" | "!=" | "eq" | "ne"
 * REL_OP  -> "<" | "<=" | ">" | ">=" | "lt" | "le" | "gt" | "ge"
 * VAR     -> a variable reference
 * IF_ELSEIF_ELSE -> an "if" statement with zero or more "elseif" components
 *  and an optional "else" component
 * FUNCTION_CALL -> FUNCTION_NAME "(" ARG_LIST ")"
 *   [no space between FUNCTION_NAME and "("]
 * ARG_LIST -> EMPTY | E2 | ARG_LIST "," E2
 *
 * A function call is a function name (an alphabetic followed by alphanumerics
 * or underscores), immediately followed by a left paren, followed by zero
 * or more comma-separated arguments (each an expression), followed by a
 * right paren.
 *
 * The names of the parsing functions correspond to the operator precedence
 * levels appearing in Table 7-3 of "C: A Reference Manual", Harbison & Steele,
 * Fourth Edition, 1995, p. 181, with a few additional operators.
 * Level 10b has a higher precedence than 10a.
 */
static int
heir(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir2(e, &res1) == -1) {
	*result = res1;
	return(-1);
  }

  if (matches(e, T_COMMA)) {
	if (heir(e, result) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "comma"));
  }
  else
	*result = res1;

  return(0);
}

static int
heir2(Lex_state *e, Expr_result *result)
{
  char *val;
  Expr_result res1;
  Lex_token *t1, *t2;
  Token op;
  Var *var;

  if (value_init(&res1, result))
	return(-1);

  /*
   * Assignments look like:  ${foo::bar} assign-op <expr>
   */
  if ((t1 = token_lookahead(e, 0)) != NULL
	  && (t2 = token_lookahead(e, 1)) != NULL
	  && t1->token == T_STRING
	  && t1->value->is_var_reference
	  && (op = match_assign_token(t2->token)) != T_UNDEF) {
	if (!matches(e, T_STRING)) {
	  seterr_s(e, result, "Error parsing assignment?");
	  return(-1);
	}
	if (!matches(e, t2->token)) {
	  seterr_s(e, result, "Error parsing assignment?");
	  return(-1);
	}
	if ((var = var_parse_name(t1->value->val.strval, NULL)) == NULL) {
	  seterr_s(e, result, "Invalid variable name on LHS of assignment");
	  return(-1);
	}
	if (var->flags != NULL) {
	  seterr_s(e, result, "No modifier flags allowed on LHS of assignment");
	  return(-1);
	}
	if (is_reserved_namespace(var->ns)) {
	  seterr_s(e, result,
			   ds_xprintf("Can't assign to invalid or reserved namespace '%s'",
						  t1->value->val.strval));
	  return(-1);
	}

	if (heir2(e, result) == -1) {
	  seterr_s(e, result, "Invalid operand");
	  return(-1);
	}

	if (!e->do_eval)
	  return(0);

	if (t2->token != T_ASSIGN) {
	  Expr_result *temp;

	  temp = init_expr_result(NULL);
	  init_value(&temp->value, T_STRING, 
				 acs_string_operand(t1->value->val.strval, e->env));

	  if (eval2(e, result, op, temp, result) == -1)
		return(-1);
	}

	if (force_string(result, &val) == -1) {
	  seterr_s(e, result, "Conversion error for RHS of assignment");
	  return(-1);
	}
	/*
	 * XXX This obviously isn't very efficient.  But we need a better
	 * mechanism for storing an Expr_result as it is rather than as
	 * a string.
	 */
	if (var_ns_new(&e->env->namespaces, var->ns, NULL) == NULL
		|| var_ns_add_key(e->env->namespaces, var->ns, var->varname, val)
		== NULL) {
	  seterr_s(e, result, ds_xprintf("Cannot assign result to \"%s\"",
									 t1->value->val.strval));
	  return(-1);
	}

	if (result->value.token == T_STRING)
	  log_msg((LOG_TRACE_LEVEL, "Set %s to \"%s\"",
			   t1->value->val.strval, acs_format_result(result)));
	else
	  log_msg((LOG_TRACE_LEVEL, "Set %s to %s",
			   t1->value->val.strval, acs_format_result(result)));
  }
  else if (matches(e, T_IF)) {
	if (do_if(e, result) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "if"));
  }
  else {
	if (heir3(e, result) == -1)
	  return(-1);
  }

  return(0);
}

/*
 * We've seen the "if" or "elseif" keyword.
 * Process the condition and if True, evaluate the statement block; if False,
 * only parse the statement block:
 *    if/elseif (cond) { if-true-block }
 * Return 1 if the condition was True, 0 if not, and -1 or error.
 */
static int
do_if_test(Lex_state *e, Expr_result *result)
{
  int eval_state, is_true, st;
  Expr_result cond, discard, *res;

  if (value_init(&cond, result))
	return(-1);

  if (!matches(e, T_LPAREN)) {
	seterr_s(e, result, "Opening paren is missing");
	return(-1);
  }

  if (heir(e, &cond) == -1)
	return(-1);

  if (!matches(e, T_RPAREN)) {
	seterr_s(e, result, "Closing paren is missing");
	return(-1);
  }

  if (!token_matches(e, 0, T_LBRACE)) {
	seterr_s(e, result, "Opening brace is missing");
	return(-1);
  }

  if (e->do_eval)
	res = result;
  else {
	if (value_init(&discard, result))
	  return(-1);
	res = &discard;
  }

  st = is_true_value(&cond.value);
  if (e->do_eval && st == -1) {
	seterr(e, result, ACS_EXPR_EVAL_ERROR, "Invalid boolean result value");
	return(-1);
  }
  if (st == 1 || !e->do_eval) {
	/* Do True block */
	if (do_block(e, res) == -1)
	  return(-1);
	is_true = 1;
  }
  else {

	/*
	 * Condition is False
	 * Parse but do not process the "if true" part.
	 */
	eval_state = e->do_eval;
	e->do_eval = 0;
	if (do_block(e, res) == -1)
	  return(-1);
	e->do_eval = eval_state;
	is_true = 0;
  }

  return(is_true);
}

/*
 * Process an if/elseif/else statement.
 * The "elseif" and "else" are optional.
 * The value of the statement is that of the last executed statement, or
 * the empty string if no statement is executed.
 */
static int
do_if(Lex_state *e, Expr_result *result)
{
  int eval_state, is_true;

  eval_state = e->do_eval;
  while (1) {
	if ((is_true = do_if_test(e, result)) == -1)
	  return(-1);

	if (is_true)
	  e->do_eval = 0;

	if (matches(e, T_ELSEIF))
	  continue;
	else if (matches(e, T_ELSE)) {
	  if (!token_matches(e, 0, T_LBRACE)) {
		seterr_s(e, result, "Opening brace is missing");
		return(-1);
	  }

	  if (e->do_eval) {
		if (do_block(e, result) == -1)
		  return(-1);
	  }
	  else {
		Expr_result discard;

		if (value_init(&discard, result))
		  return(-1);
		if (do_block(e, &discard) == -1)
		  return(-1);
	  }
	  break;
	}
	else {
	  /* There's no "else" */
	  if (e->do_eval)
		init_value(&result->value, T_STRING, "");
	  break;
	}
  }

  e->do_eval = eval_state;

  return(0);
}

static int
heir3(Lex_state *e, Expr_result *result)
{
  int eval_state, st;
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir4(e, &res1) == -1) {
	*result = res1;
	return(-1);
  }

  if (matches(e, T_COND)) {
	Expr_result discard;

	/* This is a C-style conditional expression: a ? b : c */
	if (value_init(&discard, result))
	  return(-1);

	if ((st = is_true_value(&res1.value)) == 1) {
	  if (heir(e, result) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (!matches(e, T_COLON)) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  /* Parse but don't evaluate the second expression. */
	  eval_state = e->do_eval;
	  e->do_eval = 0;
	  if (heir(e, &discard) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  e->do_eval = eval_state;
	}
	else if (e->do_eval && st == -1) {
	  seterr(e, result, ACS_EXPR_EVAL_ERROR, "Invalid boolean result value");
	  return(-1);
	}
	else {
	  /* Parse but don't evaluate the first expression. */
	  eval_state = e->do_eval;
	  e->do_eval = 0;
	  if (heir(e, &discard) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  e->do_eval = eval_state;
	  if (!matches(e, T_COLON)) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (heir(e, result) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	}
	log_msg((LOG_TRACE_LEVEL, "conditional"));
  }
  else
	  *result = res1;

  return(0);
}

/*
 * The C-style logical OR operator.
 * The tricky part is that evaluation is from left to right and stops
 * as soon as the truth or falsehood of the result is known.
 */
static int
heir4(Lex_state *e, Expr_result *result)
{
  int eval_state;
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir5(e, result) == -1)
	return(-1);

  res1 = *result;
  eval_state = e->do_eval;

  while (1) {
	if (matches(e, T_OR)) {
	  int i1, i2;

	  if (e->do_eval) {
		if (force_boolean(&res1, &i1) == -1) {
		  seterr_s(e, result, "Invalid operand");
		  return(-1);
		}
		if (i1) {
		  result->value.val.intval = 1L;
		  result->value.token = T_INTEGER;
		  e->do_eval = 0;
		  if (heir5(e, &res1) == -1) {
			seterr_s(e, result, "Invalid operand");
			return(-1);
		  }
		}
		else {
		  if (heir5(e, &res1) == -1) {
			seterr_s(e, result, "Invalid operand");
			return(-1);
		  }
		  if (force_boolean(&res1, &i2) == -1) {
			seterr_s(e, result, "Invalid operand");
			return(-1);
		  }
		  if (i2) {
			result->value.val.intval = 1L;
			result->value.token = T_INTEGER;
			e->do_eval = 0;
		  }
		  else {
			result->value.val.intval = 0L;
			result->value.token = T_INTEGER;
		  }
		}
	  }
	  else {
		if (heir5(e, &res1) == -1) {
		  seterr_s(e, result, "Invalid operand");
		  return(-1);
		}
	  }
	  log_msg((LOG_TRACE_LEVEL, "logical or"));
	}
	else
	  break;
  }

  e->do_eval = eval_state;

  return(0);
}

/*
 * The C-style logical AND operator.
 * The tricky part is that evaluation is from left to right and stops
 * as soon as the truth or falsehood of the result is known.
 */
static int
heir5(Lex_state *e, Expr_result *result)
{
  int eval_state;
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir6(e, result) == -1)
	return(-1);

  res1 = *result;
  eval_state = e->do_eval;

  while (1) {
	if (matches(e, T_AND)) {
	  int i1, i2;

	  if (e->do_eval) {
		if (force_boolean(&res1, &i1) == -1) {
		  seterr_s(e, result, "Invalid operand");
		  return(-1);
		}
		if (i1 == 0) {
		  result->value.val.intval = 0L;
		  result->value.token = T_INTEGER;
		  e->do_eval = 0;
		  if (heir5(e, &res1) == -1) {
			seterr_s(e, result, "Invalid operand");
			return(-1);
		  }
		}
		else {
		  if (heir5(e, &res1) == -1) {
			seterr_s(e, result, "Invalid operand");
			return(-1);
		  }
		  if (force_boolean(&res1, &i2) == -1) {
			seterr_s(e, result, "Invalid operand");
			return(-1);
		  }
		  if (i2 == 0) {
			result->value.val.intval = 0L;
			result->value.token = T_INTEGER;
			e->do_eval = 0;
		  }
		  else {
			result->value.val.intval = 1L;
			result->value.token = T_INTEGER;
		  }
		}
	  }
	  else {
		if (heir5(e, &res1) == -1) {
		  seterr_s(e, result, "Invalid operand");
		  return(-1);
		}
	  }
	  log_msg((LOG_TRACE_LEVEL, "logical and"));
	}
	else
	  break;
  }

  e->do_eval = eval_state;

  return(0);
}

static int
heir6(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir7(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_BITOR)) {
	  if (heir7(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_BITOR, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "bitwise or"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir7(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir8(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_BITXOR)) {
	  if (heir8(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_BITXOR, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "bit xor"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir8(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir9(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_BITAND)) {
	  if (heir9(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_BITAND, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "bit and"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir9(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir10a(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_EQ)) {
	  if (heir10a(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_EQ, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "eq"));
	}
	else if (matches(e, T_EQ_I)) {
	  if (heir10a(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_EQ_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "eq:i"));
	}
	else if (matches(e, T_NE)) {
	  if (heir10a(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_NE, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "ne"));
	}
	else if (matches(e, T_NE_I)) {
	  if (heir10a(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_NE_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "ne:i"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir10a(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir10b(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_LT)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_LT, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "lt"));
	}
	else if (matches(e, T_LT_I)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_LT_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "lt:i"));
	}
	else if (matches(e, T_LE)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_LE, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "le"));
	}
	else if (matches(e, T_LE_I)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_LE_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "le:i"));
	}
	else if (matches(e, T_EQ)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_EQ, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "eq"));
	}
	else if (matches(e, T_EQ_I)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_EQ_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "eq:i"));
	}
	else if (matches(e, T_NE)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_NE, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "ne"));
	}
	else if (matches(e, T_NE_I)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_NE_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "ne:i"));
	}
	else if (matches(e, T_GT)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_GT, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "gt"));
	}
	else if (matches(e, T_GT_I)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_GT_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "gt:i"));
	}
	else if (matches(e, T_GE)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_GE, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "ge"));
	}
	else if (matches(e, T_GE_I)) {
	  if (heir10b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_GE_I, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "ge:i"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir10b(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir11(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_DOT)) {
	  if (heir11(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_DOT, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "dot"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir11(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir12(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_BITSHL)) {
	  if (heir12(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_BITSHL, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "bit shift left"));
	}
	else if (matches(e, T_BITSHR)) {
	  if (heir12(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_BITSHR, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "bit shift right"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir12(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir13a(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_PLUS)) {
	  if (heir13a(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_PLUS, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "binary plus"));
	}
	else if (matches(e, T_MINUS)) {
	  if (heir13a(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_MINUS, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "binary minus"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir13a(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir13b(e, result) == -1)
	return(-1);

  while (1) {
	if (matches(e, T_TIMES)) {
	  if (heir13b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_TIMES, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "multiply"));
	}
	else if (matches(e, T_DIV)) {
	  if (heir13b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_DIV, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "divide"));
	}
	else if (matches(e, T_MOD)) {
	  if (heir13b(e, &res1) == -1) {
		seterr_s(e, result, "Invalid operand");
		return(-1);
	  }
	  if (eval2(e, result, T_MOD, result, &res1) == -1)
		return(-1);
	  log_msg((LOG_TRACE_LEVEL, "mod"));
	}
	else
	  break;
  }

  return(0);
}

static int
heir13b(Lex_state *e, Expr_result *result)
{
  Expr_result res1;

  if (value_init(&res1, result))
	return(-1);

  if (heir15(e, result) == -1)
	return(-1);

  if (matches(e, T_EXP)) {
	if (heir13a(e, &res1) == -1) {
	  seterr_s(e, result, "Invalid operand");
	  return(-1);
	}
	if (eval2(e, result, T_EXP, result, &res1) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "exponentiation"));
  }

  return(0);
}

static int
heir15(Lex_state *e, Expr_result *result)
{
  Expr_result res1;
  Lex_token *t;

  if (value_init(&res1, result))
	return(-1);

  if (matches(e, T_NOT)) {
	if (heir15(e, &res1) == -1) {
	  seterr_s(e, result, "Invalid operand");
	  return(-1);
	}
	if (eval1(e, result, T_NOT, &res1) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "logical not"));
  }
  else if (matches(e, T_BITCPL)) {
	if (heir15(e, &res1) == -1) {
	  seterr_s(e, result, "Invalid operand");
	  return(-1);
	}
	if (eval1(e, result, T_BITCPL, &res1) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "bit complement"));
  }
  else if (token_matches(e, 0, T_PLUS)
		   && token_matches(e, 1, T_PLUS)
		   && ((t = token_lookahead(e, 2)) != NULL
			   && t->token == T_STRING
			   && t->value->is_var_reference)) {
	/* Pre-increment */
	matches(e, T_PLUS);
	matches(e, T_PLUS);
	matches(e, T_STRING);
	if (inc_variable(e, t->value->val.strval, T_PREINC, result) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "pre-increment"));
  }
  else if (token_matches(e, 0, T_MINUS)
		   && token_matches(e, 1, T_MINUS)
		   && ((t = token_lookahead(e, 2)) != NULL
			   && t->token == T_STRING
			   && t->value->is_var_reference)) {
	/* Pre-decrement */
	matches(e, T_MINUS);
	matches(e, T_MINUS);
	matches(e, T_STRING);
	if (inc_variable(e, t->value->val.strval, T_PREDEC, result) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "pre-decrement"));
  }
  else if (matches(e, T_PLUS)) {
	if (value_init(&res1, result))
	  return(-1);
	if (heir15(e, &res1) == -1)
	  return(-1);
	if (eval1(e, result, T_UNARY_PLUS, &res1) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "unary plus"));
  }
  else if (matches(e, T_MINUS)) {
	if (value_init(&res1, result))
	  return(-1);
	if (heir15(e, &res1) == -1)
	  return(-1);
	if (eval1(e, result, T_UNARY_MINUS, &res1) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "unary minus"));
  }
  else
	return(heir17(e, result));

  return(0);
}

static int
heir17(Lex_state *e, Expr_result *result)
{
  Lex_token *t;
  Arglist *args;

  if (matches(e, T_LPAREN)) {
	if (heir(e, result) == -1)
	  return(-1);
	if (matches(e, T_RPAREN))
	  return(0);
	return(-1);
  }

  if (((t = token_lookahead(e, 0)) != NULL
	   && t->token == T_STRING
	   && t->value->is_var_reference)
	  && token_matches(e, 1, T_MINUS)
	  && token_matches(e, 2, T_MINUS)) {
	/* Post-decrement */
	matches(e, T_STRING);
	matches(e, T_MINUS);
	matches(e, T_MINUS);
	if (inc_variable(e, t->value->val.strval, T_POSTDEC, result) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "post-decrement"));
	return(0);
  }

  if (((t = token_lookahead(e, 0)) != NULL
	   && t->token == T_STRING
	   && t->value->is_var_reference)
	  && token_matches(e, 1, T_PLUS)
	  && token_matches(e, 2, T_PLUS)) {
	/* Post-increment */
	matches(e, T_STRING);
	matches(e, T_PLUS);
	matches(e, T_PLUS);
	if (inc_variable(e, t->value->val.strval, T_POSTINC, result) == -1)
	  return(-1);
	log_msg((LOG_TRACE_LEVEL, "post-increment"));
	return(0);
  }

  args = NULL;
  if ((t = matches(e, T_FUNC)) != NULL) {
	char *fname;

	fname = t->value->val.strval;
	if (funargs(e, &args) == -1) {
	  seterr_e(e, result,
			 ds_xprintf("Invalid function call argument: \"%s\"", fname));
	  return(-1);
	}
    if (eval_function(fname, e, args, result) == -1) {
	  if (result->errmsg != NULL)
		log_msg((LOG_ERROR_LEVEL, "%s", result->errmsg));
	  return(-1);
	}
	return(0);
  }

  if (is_operand(e, result)) {
	e->ctn++;
	return(0);
  }

  seterr_e(e, result, ds_xprintf("Syntax error at: \"%s\"", e->expr));
  return(-1);
}

Arglist *
expr_init_arg(void)
{
  Arglist *a;

  a = ALLOC(Arglist);
  a->result = init_expr_result(NULL);
  a->next = NULL;

  return(a);
}

/*
 * Verify the syntax of a function call, evaluating arguments as we go.
 * Build the resulting argument values into a list and return the list.
 * Return -1 if the a syntax or evaluation occurs, 0 otherwise.
 */
static int
funargs(Lex_state *e, Arglist **head)
{
  Arglist *a, *prev;

  if (!matches(e, T_LPAREN))
    return(-1);

  /* No arguments? */
  if (matches(e, T_RPAREN))
    return(0);

  /* There are one or more function arguments. */
  prev = NULL;
  while (1) {
	a = expr_init_arg();

	/*
	 * The comma operator is essentially disabled during argument list
	 * processing, so skip that production.
	 */
    if (heir2(e, a->result) == -1)
      return(-1);

    if (prev != NULL)
      prev->next = a;
    else
      *head = a;
    prev = a;

    if (!matches(e, T_COMMA))
      break;
  }

  if (!matches(e, T_RPAREN))
    return(-1);

  return(0);
}

/*
 * Validiate the function call syntax of F.
 * Convert argument types as necessary.
 * Note: conversion is done in place - the original value is lost.
 * Return -1 on failure, 0 otherwise.
 */
static int
validate_function(Lex_state *e, Func *f, Arglist *args)
{
  int n, nargs;
  char *d, *errmsg;
  Arglist *x;
  Lex_token *t;

  /*
   * Check if the correct number of arguments has been given.
   * An arity of -1 means there are a variable number of arguments.
   */
  if (f->arity != -1) {
	nargs = 0;
	for (x = args; x != NULL; x = x->next)
	  nargs++;
	if (nargs != f->arity) {
	  log_msg((LOG_ERROR_LEVEL, "Incorrect number of arguments"));
	  return(-1);
	}
  }

  if (e->do_eval == 0) {
	/*
	 * We can't examine the argument types yet because they have not been
	 * computed.
	 */
	return(0);
  }

  /*
   * Argument descriptor characters:
   * '-': leave the argument as-is
   * 's': convert the argument to a T_STRING if necessary
   * 'S': the argument must already be a T_STRING
   * 'i': convert the argument to a T_INTEGER if necessary
   * 'I': the argument must already be a T_INTEGER
   * 'r': convert the argument to a T_REAL if necessary
   * 'R': the argument must already be a T_REAL
   * '*': means zero or more arguments may follow but quit validating
   * An empty descriptor means leave all arguments as-is.
   */
  errmsg = NULL;
  n = 1;
  for (d = f->argdesc, x = args; *d != '\0' && x != NULL; d++, x = x->next) {
	switch ((int) *d) {
	case '-':
	  break;

	case '*':
	  return(0);

	case 's':
	  if (x->result->value.token == T_STRING)
		break;
	  else if (x->result->value.token == T_LITERAL) {
		x->result->value.token = T_STRING;
	  }
	  else if (x->result->value.token == T_INTEGER) {
		x->result->value.val.strval
		  = ds_xprintf("%ld", x->result->value.val.intval);
		x->result->value.token = T_STRING;
	  }
	  else if (x->result->value.token == T_REAL) {
		/* This is tricky to do properly because of printf() formatting. */
		x->result->value.val.strval
		  = ds_xprintf("%f", x->result->value.val.realval);
		x->result->value.token = T_STRING;
	  }
	  else {
		errmsg = "require a real number";
		goto bad_arg;
	  }
	  break;

	case 'S':
	  if (x->result->value.token != T_STRING
		  && x->result->value.token != T_LITERAL) {
		errmsg = "require a string or literal";
		goto bad_arg;
	  }
	  break;

	case 'i':
	  if (x->result->value.token == T_INTEGER)
		break;
	  else if (x->result->value.token == T_STRING) {
		if (is_number(x->result->value.val.strval, &t, NULL) != 1) {
		  errmsg = "require an integer number";
		  goto bad_arg;
		}
		if (t->token == T_REAL)
		  x->result->value.val.intval = (long) t->value->val.realval;
		else
		  x->result->value.val.intval = t->value->val.intval;
		x->result->value.token = T_INTEGER;
	  }
	  else if (x->result->value.token == T_REAL) {
		x->result->value.val.intval = (long) x->result->value.val.realval;
		x->result->value.token = T_INTEGER;
	  }
	  else {
		errmsg = "require an integer number";
		goto bad_arg;
	  }
	  break;

	case 'I':
	  if (x->result->value.token != T_INTEGER) {
		errmsg = "require an integer number";
		goto bad_arg;
	  }
	  break;

	case 'r':
	  if (x->result->value.token == T_REAL)
		break;
	  else if (x->result->value.token == T_INTEGER) {
		x->result->value.val.realval = (double) x->result->value.val.intval;
		x->result->value.token = T_REAL;
	  }
	  else if (x->result->value.token == T_STRING) {
		if (is_number(x->result->value.val.strval, &t, NULL) != 1) {
		  errmsg = "require a real number";
		  goto bad_arg;
		}
		if (t->token == T_REAL)
		  x->result->value.val.realval = t->value->val.realval;
		else
		  x->result->value.val.realval = (double) t->value->val.intval;
		x->result->value.token = T_REAL;
	  }
	  else {
		errmsg = "require a real number";
		goto bad_arg;
	  }
	  break;

	case 'R':
	  if (x->result->value.token != T_REAL) {
		errmsg = "require a real number";
		goto bad_arg;
	  }
	  break;

	default:
	  errmsg = "internal error";
	  goto bad_arg;
	}
	n++;
  }

  if (f->arity == -1) {
	if (x != NULL && *d == '\0') {
	  errmsg = "called with too many arguments";
	  n = 0;
	  goto bad_arg;
	}
  }
  else if (x == NULL && *d != '\0' && *d != '*') {
	n = 0;
	errmsg = "at least one required argument is missing";
	goto bad_arg;
  }

  return(0);

 bad_arg:
  if (errmsg != NULL) {
	if (n)
	  log_msg((LOG_ERROR_LEVEL, "Function \"%s\": Argument %d is invalid: %s",
			   f->name, n, errmsg));
	else
	  log_msg((LOG_ERROR_LEVEL, "Function \"%s\": %s", f->name, errmsg));
  }

  return(-1);
}

#ifdef NOTDEF
static Func *
lookup_function_name(char *name)
{
  Func *f;

  for (f = functions; f->name != NULL; f++) {
	if (streq(f->name, name))
	  return(f);
  }

  return(NULL);
}
#endif

static int
func_index(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1, *s2;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;
  x = x->next;
  s2 = x->result->value.val.strval;

  result->value.token = T_INTEGER;
  result->value.val.intval = indexset(s1, s2);

  return(0);
}

static int
func_info(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1, *s2;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;
  x = x->next;
  if (x != NULL)
	s2 = x->result->value.val.strval;
  else
	s2 = NULL;

  if (streq(s1, "namespaces") && x == NULL) {
	result->value.token = T_STRING;
	result->value.val.strval = var_ns_names(e->env->namespaces);
  }
  else if (streq(s1, "namespace") && s2 != NULL) {
	result->value.token = T_STRING;
	result->value.val.strval = var_ns_buf(e->env->namespaces, s2);
  }
  else {
	seterr_s(e, result, "Invalid function call argument to 'info'");
	return(-1);
  }

  return(0);
}

static int
func_strftime(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *buf, *s1;
  size_t rc;
  struct tm *tm;
  time_t now;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  time(&now);
  tm = localtime(&now);
  buf = (char *) malloc(1024);	/* XXX */
  if ((rc = strftime(buf, 1024, s1, tm)) == 0) {
	seterr_s(e, result,
			 ds_xprintf("strftime conversion failed for '%s'", s1));
	return(-1);
  }

  result->value.token = T_STRING;
  result->value.val.strval = buf;
  log_msg((LOG_TRACE_LEVEL, "Strftime of '%s' is %s",
		   s1, result->value.val.strval));

  return(0);
}

/*
 * Usage: strtr(<input-string>, <string1>[, <string2>[,cds]])
 *
 * String transliteration, like the tr(1) command and perl's tr and y
 * operators.
 * The first argument is the input string to be transliterated
 * (stdin in the tr command).
 * The second argument is the search list ("string1" in the tr command).
 * The third argument is the (possibly empty) replacement list
 * ("string2" in the tr command); it may be omitted if no flag string
 * argument follows.
 * The fourth, optional argument is a literal flag string made of the
 * characters 'c', 'd', and 's' (in any order), which correspond to the
 * same flags in the tr command.
 * The result is the transliterated string.
 */
static int
func_strtr(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *input, *string1, *string2, *tr_str;
  int cflag, dflag, sflag;
  Arglist *x;

  cflag = dflag = sflag = 0;

  x = arglist;
  input = x->result->value.val.strval;
  x = x->next;
  string1 = x->result->value.val.strval;
  x = x->next;
  if (x == NULL)
	string2 = NULL;
  else {
	string2 = x->result->value.val.strval;
	if (string2[0] == '\0')
	  string2 = NULL;
	x = x->next;
	/* Character flag arguments? */
	if (x != NULL) {
	  char *p;

	  p = x->result->value.val.strval;
	  if (x->next != NULL) {
		seterr_e(e, result, "strtr failed: too many arguments");
		return(-1);
	  }
	  while (*p != '\0') {
		switch ((int) *p) {
		case 'c':
		  cflag++;
		  break;
		case 'd':
		  dflag++;
		  break;
		case 's':
		  sflag++;
		  break;
		default:
		  seterr_e(e, result,
				   ds_xprintf("strtr failed: invalid flag argument: '%c'",
							  *p));
		  return(-1);
		}
		p++;
	  }
	}
  }

  log_msg((LOG_TRACE_LEVEL, "Eval: strtr(\"%s\", \"%s\")", string1, string2));
  if (strtr(input, string1, string2, cflag, dflag, sflag, &tr_str) == -1) {
	seterr_e(e, result, ds_xprintf("strtr failed: %s", tr_str));
	return(-1);
  }

  result->value.token = T_STRING;
  result->value.val.strval = tr_str;

  return(0);
}

/*
 * Return the length of the argument.
 * At present, this only applies to a string argument but it will
 * likely be useful with other kinds of objects.
 */
static int
func_length(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  result->value.token = T_INTEGER;
  result->value.val.intval = strlen(s1);
  log_msg((LOG_TRACE_LEVEL, "Length of '%s' is %d",
		   s1, result->value.val.intval));

  return(0);
}

typedef struct AttrTypeAndValue {
  char *type;
  char *value;
} AttrTypeAndValue;

/*
 * RFC 2253
 * Parse an AttrTypeAndValue element into its components.
 */
AttrTypeAndValue *
ldap_atav_parse(char *atav_str)
{
  char *p, *s;
  AttrTypeAndValue *atav;

  atav = ALLOC(AttrTypeAndValue);
  atav->type = NULL;
  atav->value = NULL;

  s = p = atav_str;
  while (*p != '\0') {
	if (*p == '=') {
	  *p++ = '\0';
	  atav->type = s;
	  atav->value = p;
	  return(atav);
	}
	p++;
  }

  /* No '=' found, invalid string */
  return(NULL);
}

/*
 * RFC 2253
 * Parse a RelativeDistinguishedName into a SEQUENCE of AttributeTypeAndValue
 * elements.
 */
Dsvec *
ldap_rdn_parse(char *rdn_str)
{
  char *p, *s;
  AttrTypeAndValue *atav;
  Dsvec *dsv;

  dsv = dsvec_init(NULL, sizeof(AttrTypeAndValue *));
  s = p = strdup(rdn_str);
  while (*p != '\0') {
	if (*p == '\\') {
	  p++;
	  if (*p == '\0') {
		/* String ends with a backslash? */
		return(NULL);
	  }
	}
	else if (*p == '+') {
	  /* End of an AttributeTypeAndValue */
	  *p++ = '\0';
	  if ((atav = ldap_atav_parse(s)) == NULL)
		return(NULL);
	  dsvec_add_ptr(dsv, atav);
	  s = p;
	}

	p++;
  }

  if (s != p) {
	if ((atav = ldap_atav_parse(s)) == NULL)
	  return(NULL);
	dsvec_add_ptr(dsv, atav);
  }

  dsvec_add_ptr(dsv, NULL);
  return(dsv);
}

/*
 * RFC 2253
 * Break the string representation of a Distinguished Name into a vector
 * (a SEQUENCE) of RelativeDistinguishedName elements, terminated by a
 * NULL element.
 * Return NULL if DN_STR is invalid, otherwise a pointer to the vector.
 */
Dsvec *
ldap_dn_parse(char *dn_str)
{
  char *p, *s;
  Dsvec *dsv;

  dsv = dsvec_init(NULL, sizeof(char *));
  s = p = strdup(dn_str);
  while (*p != '\0') {
	if (*p == '\\') {
	  p++;
	  if (*p == '\0') {
		/* String ends with a backslash? */
		return(NULL);
	  }
	}
	else if (*p == ',') {
	  /* End of an RDN */
	  *p++ = '\0';
	  if (ldap_rdn_parse(s) == NULL)	/* Check syntax */
		return(NULL);
	  dsvec_add_ptr(dsv, s);
	  s = p;
	}

	p++;
  }

  if (s != p) {
	if (ldap_rdn_parse(s) == NULL)		/* Check syntax */
	  return(NULL);
	dsvec_add_ptr(dsv, s);
  }

  dsvec_add_ptr(dsv, NULL);
  return(dsv);
}

/*
 * LDAP name manipulation functions based on RFC 2253, LDAP UTF-8 String
 * Representation of Distinguished Names
 *
 * DistinguishedName ::= RDNSequence
 *  RDNSequence ::= SEQUENCE OF RelativeDistinguishedName
 *   RelativeDistinguishedName ::= SET SIZE (1..MAX) OF AttributeTypeAndValue
 *     AttributeTypeAndValue ::= SEQUENCE {
 *       type  AttributeType,
 *       value AttributeValue }
 * o adjoining RelativeDistinguishedNames are separated by a comma character
 * o where there is a multi-valued RDN, the outputs from adjoining
 *   AttributeTypeAndValues are separated by a plus character
 * o the AttributeTypeAndValue is encoded as the string representation of
 *   the AttributeType, followed by an equals character followed by the
 *   string representation of the AttributeValue
 * o if the AttributeType is in a published table of attribute types
 *   associated with LDAP, then the type name string from that table
 *   is used, otherwise it is encoded as the dotted-decimal encoding of
 *   the AttributeType's OBJECT IDENTIFIER
 * o if the AttributeValue is of a type which does not have a string
 *   representation defined for it, then it is simply encoded as an
 *   octothorpe character followed by the hexadecimal representation of each
 *   of the bytes of the BER encoding
 * o if the UTF-8 string does not have any of the following characters
 *   which need escaping, then that string can be used as the string
 *   representation of the value
 *    - a space or "#" character occurring at the beginning of the string
 *    - a space character occurring at the end of the string
 *    - one of the characters ",", "+", """, "\", "<", ">" or ";"
 *   Implementations MAY escape other characters.
 *   If a character to be escaped is one of the list shown above, then it
 *   is prefixed by a backslash
 *   Otherwise the character to be escaped is replaced by a backslash and
 *   two hex digits, which form a single byte in the code of the character
 */
static int
func_ldap(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *func, *s1;
  Arglist *x;
  Dsvec *dsv;

  x = arglist;
  if (x->result->value.token != T_LITERAL
	  && x->result->value.token != T_STRING) {
	seterr_e(e, result, "Invalid argument to 'ldap' function");
	return(-1);
  }

  func = x->result->value.val.strval;
  x = x->next;
  s1 = x->result->value.val.strval;
  x = x->next;

  if (strcaseeq(func, "dn_index")) {
	long ind;

	if (x == NULL) {
	  seterr_e(e, result, "Missing index number argument to 'ldap' function");
	  return(-1);
	}
	if (force_integer(x->result, &ind) == -1) {
	  seterr_e(e, result, "Invalid index number argument to 'ldap' function");
	  return(-1);
	}
	x = x->next;
	if (x != NULL) {
	  seterr_e(e, result, "Extra argument to 'ldap' function");
	  return(-1);
	}
	if ((dsv = ldap_dn_parse(s1)) == NULL) {
	  seterr_e(e, result, "Invalid DN argument to 'ldap' function");
	  return(-1);
	}
	if (ind < 1) {
	  seterr_e(e, result, "Invalid index number argument to 'ldap' function");
	  return(-1);
	}
	if (ind > ((int) dsvec_len(dsv) - 1))
	  ind = dsvec_len(dsv) - 1;

	result->value.token = T_STRING;
	result->value.val.strval = dsvec_ptr(dsv, ind - 1, char *);
  }
  else if (strcaseeq(func, "dn_length")) {
	if (x != NULL) {
	  seterr_e(e, result, "Extra argument to 'ldap' function");
	  return(-1);
	}
	dsv = ldap_dn_parse(s1);
	result->value.token = T_INTEGER;
	if (dsv == NULL)
	  result->value.val.intval = -1;
	else
	  result->value.val.intval = dsvec_len(dsv) - 1;
  }
  else if (strcaseeq(func, "rdn_index")) {
	long ind;

	if (x == NULL) {
	  seterr_e(e, result, "Missing index number argument to 'ldap' function");
	  return(-1);
	}
	if (force_integer(x->result, &ind) == -1) {
	  seterr_e(e, result, "Invalid index number argument to 'ldap' function");
	  return(-1);
	}
	x = x->next;
	if (x != NULL) {
	  seterr_e(e, result, "Extra argument to 'ldap' function");
	  return(-1);
	}
	if ((dsv = ldap_rdn_parse(s1)) == NULL) {
	  seterr_e(e, result, "Invalid RDN argument to 'ldap' function");
	  return(-1);
	}
	if (ind < 1) {
	  seterr_e(e, result, "Invalid index number argument to 'ldap' function");
	  return(-1);
	}
	if (ind > ((int) dsvec_len(dsv) - 1))
	  ind = dsvec_len(dsv) - 1;

	result->value.token = T_STRING;
	result->value.val.strval = dsvec_ptr(dsv, ind - 1, char *);
  }
  else if (strcaseeq(func, "rdn_attrtype")
		   || strcaseeq(func, "rdn_attrvalue")) {
	long ind;
	AttrTypeAndValue *atav;

	if (x != NULL) {
	  if (force_integer(x->result, &ind) == -1) {
		seterr_e(e, result,
				 "Invalid index number argument to 'ldap' function");
		return(-1);
	  }
	  x = x->next;
	  if (x != NULL) {
		seterr_e(e, result, "Extra argument to 'ldap' function");
		return(-1);
	  }
	}
	else
	  ind = 1;

	if ((dsv = ldap_rdn_parse(s1)) == NULL) {
	  seterr_e(e, result, "Invalid RDN argument to 'ldap' function");
	  return(-1);
	}
	if (ind < 1) {
	  seterr_e(e, result, "Invalid index number argument to 'ldap' function");
	  return(-1);
	}
	if (ind > ((int) dsvec_len(dsv) - 1))
	  ind = dsvec_len(dsv) - 1;
	atav = dsvec_ptr(dsv, ind - 1, AttrTypeAndValue *);
	result->value.token = T_STRING;
	if (strcaseeq(func, "rdn_attrtype"))
	  result->value.val.strval = strdup(atav->type);
	else
	  result->value.val.strval = strdup(atav->value);
  }
  else if (strcaseeq(func, "rdn_length")) {
	if (x != NULL) {
	  seterr_e(e, result, "Extra argument to 'ldap' function");
	  return(-1);
	}
	dsv = ldap_rdn_parse(s1);
	result->value.token = T_INTEGER;
	if (dsv == NULL)
	  result->value.val.intval = -1;
	else
	  result->value.val.intval = dsvec_len(dsv) - 1;
  }
  else {
	seterr_e(e, result, "Unrecognized mode argument to 'ldap' function");
	return(-1);
  }

  return(0);
}

static int
func_mime_decode(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1;
  unsigned char *decoded;
  long nbytes;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  if ((nbytes = mime_decode_base64(s1, &decoded)) == -1) {
	seterr_e(e, result, "Base64 decoding failed");
	return(-1);
  }

  /* XXX What if it's a binary string?  Maybe we need a new Token type... */
  if (memchr(decoded, (int) '\0', (size_t) nbytes - 2) != NULL) {
	seterr_e(e, result, "Base64 decoding yielded a binary value");
	return(-1);
  }

  result->value.token = T_STRING;
  result->value.val.strval = (char *) decoded;

  return(0);
}

static int
func_mime_encode(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  unsigned char *s1;
  size_t len;
  Arglist *x;

  x = arglist;
  len = strlen(x->result->value.val.strval);
  s1 = (unsigned char *) x->result->value.val.strval;

  result->value.token = T_STRING;
  mime_encode_base64(s1, len, &result->value.val.strval);

  return(0);
}

/*
 * Usage: regsub(<source>, <regex>, <replacement>[,nocase][,repeat])
 *
 * Match REGEX against SOURCE and return the string that results when
 * substitutions are applied to DEST.
 * This is similar to the ed/ex/vi command s/REGEX/REPLACEMENT/, where
 * the current line is SOURCE.
 * Examples:
 *   regsub("hello world", "world", "auggie") -> "hello auggie"
 *   regsub("hello", ".*", "& &")             -> "hello hello"
 *   regsub("one two three", "(.*) (.*) (.*)", "\${3} \${2} \${1}")
 */
static int
func_regsub(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int cflags, repeat;
  char *errmsg, *p, *src_string, *repl_string, *regex_string;
  Arglist *x;

  x = arglist;
  src_string = x->result->value.val.strval;
  x = x->next;
  regex_string = x->result->value.val.strval;
  x = x->next;
  repl_string = x->result->value.val.strval;
  x = x->next;

  cflags = REG_EXTENDED;
  repeat = 0;

  while (x != NULL) {
	if (x->result->value.token == T_LITERAL
		&& strcaseeq(x->result->value.val.strval, "nocase"))
	  cflags |= REG_ICASE;
	else if (x->result->value.token == T_LITERAL
			 && strcaseeq(x->result->value.val.strval, "repeat"))
	  repeat = 1;
	else {
	  seterr_e(e, result, "Invalid argument to 'regsub' function");
	  return(-1);
	}
	x = x->next;
  }

  if ((p = strregexsub(src_string, regex_string, repl_string, cflags,
					   repeat, &errmsg)) == NULL) {
	log_msg((LOG_ERROR_LEVEL, "strregexsub: %s", errmsg));
	seterr_s(e, result,
			 ds_xprintf("Error occurred in 'regsub', src='%s', regex='%s'",
						src_string, regex_string));
	return(-1);
  }

  result->value.token = T_STRING;
  result->value.val.strval = p;

  return(0);
}

/*
 * Usage: match(<string>, <regex>[,namespace][,nocase])
 *
 * The regular expression REGEX is matched with STRING,
 * with a '^' (the start-of-string anchor) implicitly prepended.
 * REGEX may contain subexpressions enclosed between '(' and ')'
 * (or '\(' and '\)').
 * If the optional 'nocase' literal argument is given, then matching is done
 * case-insensitively. Only one parenthesized pair can be used.
 * IEEE Std 1003.2 ("POSIX.2") "extended" regular expressions are used.
 *
 * If the match fails, the result is 0. If the match succeeds there are
 * several possibilities:
 *
 * - if there are no subexpressions in REGEX, the result is an integer that
 * is the number of characters matched.
 *
 * - if there is at least one subexpression in REGEX but no namespace (a
 * string argument) is given, the result is the substring of STRING that was
 * matched by the entire regular expression.
 *
 * - if there is at least one subexpression in REGEX and a namespace argument
 * is given, the value of the first matching subexpression is assigned to the
 * variable "1" of the namespace, the value of the second subexpression is
 * assigned to the variable "2" of the namespace, and so on up to the ninth
 * subexpression. Variable "0" of the namespace is assigned the substring of
 * STRING that was matched by the entire regular expression.
 */
static int
func_regmatch(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int icase, st;
  char *errmsg, *regex, *str, *ns;
  Arglist *x;
  Dsvec *matches;

  x = arglist;
  str = x->result->value.val.strval;
  x = x->next;
  regex = ds_xprintf("^%s", x->result->value.val.strval);
  x = x->next;
  if (x != NULL && x->result->value.token == T_STRING) {
	ns = x->result->value.val.strval;
	if (is_reserved_namespace(ns)) {
	  seterr_s(e, result,
			   ds_xprintf("Can't assign to reserved namespace '%s'", ns));
	  return(-1);
	}
	x = x->next;
  }
  else
	ns = NULL;

  if (x != NULL) {
	if (x->result->value.token != T_LITERAL
		|| !strcaseeq(x->result->value.val.strval, "nocase")
		|| x->next != NULL) {
	  seterr_e(e, result, "Invalid argument to 'regmatch' function");
	  return(-1);
	}
	icase = REG_ICASE;
  }
  else
	icase = 0;

  matches = dsvec_init(NULL, sizeof(char *));
  st = strregex(str, regex, NULL, REG_EXTENDED | icase, matches, &errmsg);
  if (st == -1) {
	seterr_e(e, result, errmsg);
	return(-1);
  }
  else if (st == 0) {
	/* No match */
	result->value.token = T_INTEGER;
	result->value.val.intval = 0;
  }
  else {
	str = (char *) dsvec_ptr_index(matches, 0);

	if (dsvec_len(matches) == 1) {
	  /* The regexp didn't have any substring matches. */
	  result->value.token = T_INTEGER;
	  result->value.val.intval = strlen(str);
	}
	else if (ns == NULL) {
	  result->value.token = T_STRING;
	  result->value.val.strval = str;
	}
	else {
	  unsigned int ui;
	  Kwv *kwv;
	  Var_ns *vs;

	  vs = var_ns_new(&e->env->namespaces, ns, NULL);
	  vs->temporary = 1;
	  kwv = vs->kwv;

	  kwv_replace(kwv, "0", str);
	  result->value.token = T_INTEGER;
	  result->value.val.intval = strlen(str);

	  for (ui = 0; ui < dsvec_len(matches); ui++) {
		str = (char *) dsvec_ptr_index(matches, ui);
		kwv_replace(kwv, ds_xprintf("%d", ui), str);
	  }
	}
  }

  return(0);
}

static int
func_substr(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int n;
  char *s1;
  long i1, i2, len;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;
  x = x->next;
  i1 = x->result->value.val.intval;
  x = x->next;
  i2 = x->result->value.val.intval;

  if (i1 <= 0) {
	seterr_e(e, result, "Second arg to substr must be greater than zero");
	return(-1);
  }
  if (i2 <= 0) {
	seterr_e(e, result, "Third arg to substr must be greater than zero");
	return(-1);
  }

  result->value.token = T_STRING;
  len = strlen(s1);
  if (i1 > len)
	result->value.val.strval = "";
  else {
	char *start;
	Ds ds;

	start = s1 + (int) i1 - 1;
	n = len - (int) i1 + 1;
	n = (int) i2 > n ? n : (int) i2;
	ds_init(&ds);
	ds_concatn(&ds, start, n);
	result->value.val.strval = ds_buf(&ds);
  }

  return(0);
}

/*
 * Decode the URL-style %XX encoded string.
 * RFC 2396 URI Generic Syntax
 */
static int
func_url_decode(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *decoded, *s1;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  if ((decoded = url_decode(s1, NULL, NULL)) == NULL) {
	seterr_s(e, result, ds_xprintf("url_decode failed for '%s'", s1));
	return(-1);
  }

  result->value.token = T_STRING;
  result->value.val.strval = decoded;

  return(0);
}

/*
 * URL encode the argument.
 * RFC 2396 URI Generic Syntax
 */
static int
func_url_encode(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  result->value.token = T_STRING;
  result->value.val.strval = url_encode(s1);

  return(0);
}

/*
 * Evaluate the argument to the user() predicate and user_revoke() function
 */

typedef struct User_expr {
  char *expr;
  Dsvec *dsv;
  int current;
  Credentials *credentials;
  char *errmsg;
} User_expr;

typedef struct User_token {
  Token token;
  char *value;
} User_token;

static int ue_filter1(User_expr *expr, int *result);
static int ue_filter2(User_expr *expr, int *result);
static int ue_filter3(User_expr *expr, int *result);
static int ue_filter4(User_expr *expr, int *result);

/* For debugging */
static MAYBE_UNUSED void
ue_dump(User_expr *expr)
{
  int i;
  User_token *ut;

  for (i = 0; i < dsvec_len(expr->dsv); i++) {
	ut = dsvec_ptr(expr->dsv, i, User_token *);

	switch (ut->token) {
	case T_OR:
	  fprintf(stderr, "or\n");
	  break;
	case T_AND:
	  fprintf(stderr, "and\n");
	  break;
	case T_LPAREN:
	  fprintf(stderr, "(\n");
	  break;
	case T_RPAREN:
	  fprintf(stderr, ")\n");
	  break;
	case T_NOT:
	  fprintf(stderr, "!\n");
	  break;
	case T_EOI:
	  fprintf(stderr, "EOI\n");
	  break;
	case T_STRING:
	  fprintf(stderr, "%s\n", ut->value);
	  break;
	default:
	  fprintf(stderr, "???\n");
	  break;
	}
  }
}

/*
 * Look for the next token, skipping initial whitespace.
 * Set PTR to the character following the token.
 *
 * Tokens are:
 * o ||, or
 * o &&, and
 * o (, )
 * o !, not
 * o anything else is a "string token"
 */
static User_token *
ue_lex(char *str, char **ptr)
{
  size_t len;
  char *p, *s;
  User_token *ut;

  p = str;
  while (*p == ' ' || *p == '\t')
	p++;
  s = p;

  ut = ALLOC(User_token);
  if (*s == '\0') {
	ut->token = T_EOI;
	ut->value = NULL;
	len = 0;
  }
  else if (strneq(s, "or", 2) && !isalpha((int) *(s + 2))) {
	ut->token = T_OR;
	ut->value = NULL;
	len = 2;
  }
  else if (*s == '|' && *(s + 1) == '|') {
	ut->token = T_OR;
	ut->value = NULL;
	len = 2;
  }
  else if (strneq(s, "and", 3) && !isalpha((int) *(s + 3))) {
	ut->token = T_AND;
	ut->value = NULL;
	len = 3;
  }
  else if (*s == '&' && *(s + 1) == '&') {
	ut->token = T_AND;
	ut->value = NULL;
	len = 2;
  }
  else if (*s == '(') {
	ut->token = T_LPAREN;
	ut->value = NULL;
	len = 1;
  }
  else if (*s == '!') {
	ut->token = T_NOT;
	ut->value = NULL;
	len = 1;
  }
  else if (strneq(s, "not", 3) && !isalpha((int) *(s + 3))) {
	ut->token = T_NOT;
	ut->value = NULL;
	len = 3;
  }
  else if (*s == ')') {
	ut->token = T_RPAREN;
	ut->value = NULL;
	len = 1;
  }
  else {
	static char *break_char = " \t|&(!)";

	while (*p != '\0' && strchr(break_char, (int) *p) == NULL)
	  p++;
	ut->token = T_STRING;
	len = p - s;
	ut->value = strndup(s, len);
  }

  *ptr = s + len;

  return(ut);
}

/*
 * Lexical scanner
 */
static Dsvec *
ue_lex_scan(char *expr)
{
  char *nextp;
  Dsvec *dsv;
  User_token *ut;  

  dsv = dsvec_init(NULL, sizeof(User_token));

  do {
	ut = ue_lex(expr, &nextp);
	dsvec_add_ptr(dsv, (void *) ut);
	expr = nextp;
  } while (ut->token != T_EOI);

  return(dsv);
}

static User_token *
ue_tcurrent(User_expr *expr)
{
  User_token *ut;

  ut = dsvec_ptr(expr->dsv, expr->current, User_token *);

  return(ut);
}

static User_token *
ue_tnext(User_expr *expr)
{
  User_token *ut;

  ut = dsvec_ptr(expr->dsv, expr->current + 1, User_token *);

  return(ut);
}

static int
ue_tmatches(User_expr *expr, Token token)
{
  User_token *ut;

  ut = dsvec_ptr(expr->dsv, expr->current, User_token *);

  if (ut->token == token) {
	expr->current++;
	return(1);
  }

  return(0);
}

static MAYBE_UNUSED Token
ue_tpeek(User_expr *expr)
{
  User_token *ut;

  ut = dsvec_ptr(expr->dsv, expr->current, User_token *);

  if (ut == NULL)
	return(T_UNDEF);

  return(ut->token);
}

/*
 * Evaluate a primary term
 */
static int
ue_eval(User_expr *expr, int *result)
{
  int st;
  char *arg, *op, *p;
  Credentials *cr;
  User_token *ut;

  if ((ut = ue_tcurrent(expr)) == NULL)
	return(-1);

  if (ut->token != T_STRING)
	return(-1);

  op = ut->value;
  cr = expr->credentials;

  if ((p = strcaseprefix(op, "version")) != NULL) {
	expr->current++;
	ut = ue_tcurrent(expr);
	arg = ut->value;
	log_msg((LOG_TRACE_LEVEL, "Check for version \"%s\"", arg));
	st = (cr != NULL && streq(arg, cr->version));
  }
  else if ((p = strcaseprefix(op, "style")) != NULL) {
	expr->current++;
	ut = ue_tcurrent(expr);
	arg = ut->value;
	log_msg((LOG_TRACE_LEVEL, "Check for style \"%s\"", arg));
	if (strncaseprefix(arg, "password", 4))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_PASSWORD));
	else if (strncaseprefix(arg, "certificate", 4))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_CERT));
	else if (strncaseprefix(arg, "prompted", 6))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_PROMPTED));
	else if (strncaseprefix(arg, "native", 3))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_NATIVE));
	else if (strcaseeq(arg, "expr"))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_EXPR));
	else if (strcaseeq(arg, "admin"))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_ADMIN));
	else if (strncaseprefix(arg, "imported", 6))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_IMPORTED));
	else if (strcaseeq(arg, "alien"))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_ALIEN));
	else if (strncaseprefix(arg, "generated", 3))
	  st = (cr != NULL && (cr->auth_style & AUTH_STYLE_GENERATED));
	else
	  return(-1);
  }
  else if ((p = strcaseprefix(op, "imported_by ")) != NULL) {
	expr->current++;
	ut = ue_tcurrent(expr);
	arg = ut->value;
	log_msg((LOG_TRACE_LEVEL, "Check for importedby \"%s\"", arg));
	st = (cr->imported_by != NULL && streq(cr->imported_by, arg));
  }
  else {
	User_token *ut_next;
	DACS_name_cmp m;

	if ((ut_next = ue_tnext(expr)) != NULL && ut_next->token == T_STRING
		&& (m = lookup_name_cmp(ut_next->value)) != DACS_NAME_CMP_UNKNOWN)
	  expr->current++;
	else
	  m = get_name_cmp_mode();

	if ((st = is_matching_user_identity(op, cr, m, &expr->errmsg)) == -1)
	  return(-1);
  }

  expr->current++;
  *result = st;

  return(0);
}

/*
 * Lowest priority
 */
static int
ue_filter1(User_expr *expr, int *result)
{
  int res1, res2;

  res1 = 0;
  if (ue_filter2(expr, &res1) == -1)
	return(-1);

  while (1) {
	if (ue_tmatches(expr, T_OR)) {
	  res2 = 0;
	  if (ue_filter2(expr, &res2) == -1)
		return(-1);

	  res1 = res1 || res2;
	}
	else
	  break;
  }

  *result = res1;

  return(0);
}

static int
ue_filter2(User_expr *expr, int *result)
{
  int res1, res2;

  res1 = 0;
  if (ue_filter3(expr, &res1) == -1)
	return(-1);

  while (1) {
	if (ue_tmatches(expr, T_AND)) {
	  res2 = 0;
	  if (ue_filter2(expr, &res2) == -1)
		return(-1);
	  res1 = res1 && res2;
	}
	else
	  break;
  }

  *result = res1;

  return(0);
}

static int
ue_filter3(User_expr *expr, int *result)
{
  int res;
  
  if (ue_tmatches(expr, T_NOT)) {
	res = 0;
	if (ue_filter3(expr, &res) == -1)
	  return(-1);
	*result = !res;
  }
  else
	return(ue_filter4(expr, result));

  return(0);
}

static int
ue_filter4(User_expr *expr, int *result)
{
  
  if (ue_tmatches(expr, T_LPAREN)) {
	if (ue_filter1(expr, result) == -1)
	  return(-1);
	if (ue_tmatches(expr, T_RPAREN))
	  return(0);
	return(-1);
  }

  if (ue_eval(expr, result) == -1)
	return(-1);

  return(0);
}

/*
 * Execute the user filter expression EXPR for each set of credentials
 * (or the unauthenticated user), optionally returning the list of matches.
 * Return the number of matches, or -1 if an error occurs.
 *
 * Grammar:
 *
 * EXP -> E1
 * E1  -> E2 | E2 <or> E2
 * E2  -> E3 | E3 <and> E2
 * E3  -> E4 | <not> E3
 * E4  -> primary | '(' E1 ')'
 */
static int
ue_exec(char *expr, Credentials *credentials, Dsvec **matched, char **errmsg)
{
  int n, result;
  Credentials *cr;
  User_expr ue;

  log_msg((LOG_TRACE_LEVEL, "ue_exec: \"%s\"", expr));

  ue.expr = expr;
  ue.dsv = ue_lex_scan(expr);
  ue.errmsg = NULL;

  cr = credentials;
  n = 0;

  do {
	ue.credentials = cr;
	ue.current = 0;

	result = 0;
	if (ue_filter1(&ue, &result) != 0) {
	  *errmsg = ue.errmsg;
	  return(-1);
	}

	if (!ue_tmatches(&ue, T_EOI)) {
	  *errmsg = "Syntax error";
	  return(-1);
	}

	if (result) {
	  n++;
	  if (matched != NULL) {
		if (*matched == NULL)
		  *matched = dsvec_init(NULL, sizeof(Credentials *));
		dsvec_add_ptr(*matched, cr);
	  }
	}

	if (cr != NULL)
	  cr = cr->next;
  } while (cr != NULL);
	
  return(n);
}

/*
 * User identity predicate.
 * The single string argument is a filter expression that must be
 * satisfied by a single set of credentials (or the unauthenticated user).
 * Return TRUE iff the expression is satisfied.
 * All valid credentials are considered.
 */
static int
func_user(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int st;
  char *errmsg, *s1;
  Arglist *x;

  x = arglist;
  if (x == NULL
	  || x->next != NULL
	  || (x->result->value.token != T_STRING
		  && x->result->value.token != T_LITERAL)
	  || ((s1 = x->result->value.val.strval) == NULL)
	  || *s1 == '\0') {
	seterr_e(e, result, "Invalid argument to 'user' function");
	return(-1);
  }

  if ((st = ue_exec(s1, e->env->credentials, NULL, &errmsg)) == -1) {
	if (errmsg != NULL)
	  seterr_e(e, result, ds_xprintf("Error in 'user' function: %s", errmsg));
	return(-1);
  }

  result->value.token = T_INTEGER;
  result->value.val.intval = st;

  return(0);
}

static int
func_user_revoke(Lex_state *e, Arglist *arglist, Expr_result *result)
{
#ifndef NOTDEF
  seterr_e(e, result, "Function 'revoke' is not implemented yet");
  return(-1);
#else
  int st;
  char *errmsg, *s1;
  Arglist *x;
  Dsvec *matched;

  x = arglist;
  if (x == NULL
	  || x->next != NULL
	  || (x->result->value.token != T_STRING
		  && x->result->value.token != T_LITERAL)
	  || ((s1 = x->result->value.val.strval) == NULL)
	  || *s1 == '\0') {
	seterr_e(e, result, "Invalid argument to 'revoke' function");
	return(-1);
  }

  /*
   * XXX special case to reset? arg eq.i "none"
   */

  matched = NULL;
  if ((st = ue_exec(s1, e->env->credentials, &matched, &errmsg)) == -1) {
	if (errmsg != NULL)
	  seterr_e(e, result, ds_xprintf("Error in 'ignore' function: %s",
									 errmsg));
	return(-1);
  }

  /*
   * If the unauthenticated user was matched it means "deny".
   * Otherwise, each matching identity should be ignored
   * (this could be scope-limited, but only if useful).
   */

  result->value.token = T_INTEGER;
  result->value.val.intval = st;

  return(0);
#endif
}

/*
 * Go through all credentials searching for a DACS admin identity.
 * Return TRUE iff one is found, 0 otherwise.
 */
static int
func_dacs_admin(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  Arglist *x;

  x = arglist;

  if (x != NULL) {
	seterr_e(e, result, "Invalid argument to 'dacs_admin' function");
	return(-1);
  }

  result->value.token = T_INTEGER;
  result->value.val.intval = 0;
  if (e->env->credentials == NULL)
	log_msg((LOG_TRACE_LEVEL, "No credentials, no admin identity match"));
  else if (is_dacs_admin(e->env->credentials))
	result->value.val.intval = 1;

  return(0);
}

static Dsvec *
char_sep_list_parser(char *list_format, char *spec)
{
  char *decoded, *el, *p, *s;
  Dsvec *v;

  if (spec == NULL)
	return(NULL);

  v = dsvec_init(NULL, sizeof(char *));
  s = p = strdup(spec);

  while ((el = strsep(&p, list_format)) != NULL) {
	if (*el != '\0') {
	  if ((decoded = url_decode(el, NULL, NULL)) == NULL)
		return(NULL);
	  dsvec_add_ptr(v, decoded);
	}
  }

  return(v);
}

/*
 * Return 1 if each element of SUBLIST appears in SUPERLIST,
 * 0 if an element not in SUPERLIST is found, and -1 on error.
 * Do case insensitive comparisons if IGNORE_CASE is non-zero.
 */
static int
check_subset(Dsvec *subset, Dsvec *superset, int icase)
{
  int i, j;
  char *p, *q;

  i = 0;
  while ((p = dsvec_ptr(subset, i, char *)) != NULL) {
	j = 0;
	while ((q = dsvec_ptr(superset, j, char *)) != NULL) {
	  if (icase ? strcaseeq(p, q) : streq(p, q))
		break;

	  j++;
	}
	if (q == NULL)
	  return(0);
	i++;
  }

  return(1);
}

/*
 * Usage:
 * subset(<list-format-identifier>, <purported-subset>, <superset>[, nocase])
 *
 * The <list-format-identifier> tells us how to parse the following
 * arguments.  If the format is a single punctuation character,
 * then that character separates the elements.
 * <purported-subset> is the list whose elements are
 * each required to appear somewhere in the list <superset>.
 *
 * Example:
 *  <allow>
 *    ${REQUEST:Pi} eq "GetMap" and
 *    subset(",", ${LAYERS:Pi}, "RELIEF:Foundation,GTOPO30:Foundation")
 * </allow>
 *
 * This says that the REQUEST parameter (case insensitive) must be "GetMap"
 * and that the elements of the LAYERS parameter (case insensitive) must
 * exactly match either RELIEF:Foundation or GTOPO30:Foundation; if an
 * element appears in LAYERS that does not also appear in the given list, the
 * predicate is False, otherwise it is True.
 */
static int
func_subset(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int icase, rc;
  char *s1, *s2, *t;
  Arglist *x;
  Dsvec *subset, *superset;
  Dsvec *(*parser)(char *, char *);

  x = arglist;
  t = x->result->value.val.strval;
  x = x->next;
  s1 = x->result->value.val.strval;
  x = x->next;
  s2 = x->result->value.val.strval;
  x = x->next;
  if (x != NULL) {
	if (x->result->value.token != T_LITERAL
		|| !strcaseeq(x->result->value.val.strval, "nocase")
		|| x->next != NULL) {
	  seterr_e(e, result, "Invalid argument to 'subset' function");
	  return(-1);
	}
	icase = 1;
  }
  else
	icase = 0;

  parser = NULL;
  if (ispunct((int) *t) && *(t + 1) == '\0')
	parser = char_sep_list_parser;
  else {
	seterr_e(e, result, "Unrecognized format type for 'subset' function");
	return(-1);
  }

  if ((subset = parser(t, s1)) == NULL) {
	seterr_e(e, result,
			 "Parse failed on subset argument for 'subset' function");
	return(-1);
  }

  if ((superset = parser(t, s2)) == NULL) {
	seterr_e(e, result,
		   "Parse failed on superset argument for 'subset' function");
	return(-1);
  }

  rc = check_subset(subset, superset, icase);
  if (rc == -1) {
	seterr_e(e, result, "Evaluation of 'subset' function failed");
	return(-1);
  }

  result->value.token = T_INTEGER;
  result->value.val.intval = rc;

  return(0);
}

/*
 * Return the number of elements of TESTSET that appear at least once
 * in TARGETSET.
 */
static int
contains_count(Dsvec *testset, Dsvec *targetset, int icase)
{
  int i, j, n;
  char *p, *q;

  i = 0;
  n = 0;
  while ((p = dsvec_ptr(testset, i, char *)) != NULL) {
	j = 0;
	while ((q = dsvec_ptr(targetset, j, char *)) != NULL) {
	  if (icase ? strcaseeq(p, q) : streq(p, q)) {
		n++;
		break;
	  }
	  j++;
	}
	i++;
  }

  return(n);
}

/*
 * Function to test if acknowledgements have been obtained.
 * If so, execution continues; if not, an exception is triggered
 * (ACS_DENIAL_REASON_ACK_NEEDED).
 * Usages:
 *    1. ack(notice-list)
 *           -- require an ack that matches CURRENT_URI_NO_QUERY
 *    2. ack(notice-list, EXACT_MATCH[, uri-list])
 *           -- basically the same as #1, except for additional arguments.
 *    3. ack(notice-list, ALL_ACKS[, uri-list])
 *           -- require that the set of all acks collectively name
 *              all elements of notice-list
 * Not implemented:
 *    4. ack(notice-list, ANY_ACK[, uri-list])
 *           -- require a single ack that names both a resource in uri-list
 *              and all elements of notice-list
 *
 * Each URI in notice-list identifies a notice that must be acknowledged.
 * Each URI in uri-list identifies a resource that may have been acknowledged.
 * Both notice-list and uri-list are comma-separated lists of URI
 * (optionally, with spaces following a comma).
 *
 * Additionally, the argument(s) [, NEW_ACK[, uri-list]] may appear as the
 * last arguments; NEW_ACK indicates that a new acknowlegement should be
 * issued and uri-list, if present, indicates the URIs that this ack
 * acknowledges instead of the default, CURRENT_URI_NO_QUERY.
 */
static int
func_ack(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  unsigned int ui;
  char *mode, *s1;
  Arglist *x;
  Dsvec *needed_notices, *notices;
  extern Dsvec *get_needed_notices(Dsvec *, char *mode);

  x = arglist;
  if (x == NULL) {
	seterr_e(e, result, "Invalid argument to 'ack' function");
	return(-1);
  }

  notices = dsvec_init(NULL, sizeof(char *));
  mode = "EXACT_MATCH";

  for (x = arglist; x != NULL; x = x->next) {
	if (x->result->value.token != T_STRING
		&& x->result->value.token != T_LITERAL) {
	  seterr_e(e, result, "Invalid argument to 'ack' function");
	  return(-1);
	}

	s1 = x->result->value.val.strval;
	if (strcaseeq(s1, "EXACT_MATCH") || strcaseeq(s1, "ALL_MATCH")) {
	  mode = s1;
	  x = x->next;
	  break;
	}

	if (uri_parse(s1) == NULL) {
	  seterr_e(e, result, "Invalid URI argument to 'ack' function");
	  return(-1);
	}
	dsvec_add_ptr(notices, s1);
  }

  if (x != NULL) {
	seterr_e(e, result, "Invalid argument to 'ack' function");
	return(-1);
  }

  /*
   * Check which of the required acknowledgements for this service request
   * have already been obtained and which ones are still required.
   * If none are required, this predicate should return True; otherwise,
   * access should be denied, rule processing terminated immediately, and
   * an ACS_DENIAL_REASON_ACK_NEEDED exception triggered.
   * If no acknowledgements have been collected for this service request,
   * a new NAT may be issued later; it may name the service request, a
   * caller-specified list of URI, or a list of URI obtained from the
   * matching rule.
   * If a NAT applies to this service request but one or more acknowledgements
   * are still needed, then this NAT may be replaced later.
   */
  for (ui = 0; ui < dsvec_len(notices); ui++) {
	log_msg((LOG_TRACE_LEVEL, "Rule wants ack for notice: %s",
			 (char *) dsvec_ptr_index(notices, ui)));
  }

  if ((needed_notices = get_needed_notices(notices, mode)) != NULL) {
	log_msg((LOG_TRACE_LEVEL, "Notices not yet acknowledged: %s",
			 strjoin(needed_notices, " ")));
	e->env->notices = needed_notices;
	longjmp(jmp_env, 2);
  }
  else {
	log_msg((LOG_TRACE_LEVEL, "No acks needed"));
	result->value.token = T_INTEGER;
	result->value.val.intval = 1;
  }

  return(0);
}

/*
 * Usage:
 * contains_any(<list-format-identifier>, <test-set>, <target-set>[, nocase])
 *
 * The <list-format-identifier> tells us how to parse the following
 * arguments. The result is the count of the number of elements of <test-set>
 * that appear in <target-set> at least once.
 *
 * Example:
 *    contains_any(",", ${LAYERS:Pi}, "Nests,Secret_roads,Cultural_heritage")
 *
 * Here, the value of the expression is two if any two elements in the
 * LAYERS parameter (case insensitive) appear in the given list, otherwise
 * the value is zero.
 */
static int
func_contains_any(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int icase, rc;
  char *s1, *s2, *t;
  Arglist *x;
  Dsvec *subset, *superset;
  Dsvec *(*parser)(char *, char *);

  x = arglist;
  t = x->result->value.val.strval;
  x = x->next;
  s1 = x->result->value.val.strval;
  x = x->next;
  s2 = x->result->value.val.strval;
  x = x->next;
  if (x != NULL) {
	if (x->result->value.token != T_LITERAL
		|| !strcaseeq(x->result->value.val.strval, "nocase")
		|| x->next != NULL) {
	  seterr_e(e, result, "Invalid argument to 'contains_any' function");
	  return(-1);
	}
	icase = REG_ICASE;
  }
  else
	icase = 0;

  parser = NULL;
  if (ispunct((int) *t) && *(t + 1) == '\0')
	parser = char_sep_list_parser;
  else {
	seterr_e(e, result,
			 "Unrecognized format type for 'contains_any' function");
	return(-1);
  }

  if ((subset = parser(t, s1)) == NULL) {
	seterr_e(e, result,
		   "Parse failed on subset argument for 'contains_any' function");
	return(-1);
  }

  if ((superset = parser(t, s2)) == NULL) {
	seterr_e(e, result,
		   "Parse failed on superset argument for 'contains_any' function");
	return(-1);
  }

  rc = contains_count(subset, superset, icase);
  if (rc == -1) {
	seterr_e(e, result, "Evaluation of 'contains_any' function failed");
	return(-1);
  }

  result->value.token = T_INTEGER;
  result->value.val.intval = rc;

  return(0);
}

/*
 * Re-evaluate the argument as an expression.
 * Example: eval("strlen(\"abc\")") is 3
 */
static int
func_eval(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  return(func_expr(s1, e->env, result));
}

/*
 * Stop evaluating (but continue parsing) and return the argument as
 * the final result.
 */
static int
func_exit(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  Arglist *x;

  x = arglist;
  copy_value(&result->value, &x->result->value);
  result->err = x->result->err;
  result->errmsg = x->result->errmsg;
  e->do_eval = 0;

  return(0);
}

/*
 * This function always returns its argument (a string expected to have a
 * form acceptable as the second argument to a DACS error handler directive)
 * and sets the redirect flag; it is up to the caller to interpret this.
 * Evaluation stops but parsing continues.
 * In the context of a Deny clause, the intent is to deny access and redirect
 * the user to the given URL.
 */
static int
func_redirect(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  if (acs_parse_error_handler(NULL, s1) == NULL) {
	seterr_e(e, result, "Invalid argument to 'redirect' function");
	return(-1);
  }

  e->env->redirect = s1;
  log_msg((LOG_DEBUG_LEVEL, "Request to redirect to \"%s\"", s1));
  longjmp(jmp_env, 1);

  /*NOTREACHED*/
  return(-1);
}

static int
func_file(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int argc, st;
  char *s1;
  Arglist *x;
  Dsvec *argv;
  extern int file_func(int argc, char **argv, char **result);

  argv = dsvec_init(NULL, sizeof(char *));
  argc = 0;
  for (x = arglist; x != NULL; x = x->next) {
	if (x->result->value.token != T_STRING
		&& x->result->value.token != T_LITERAL) {
	  seterr_e(e, result, "Invalid argument to 'file' function");
	  return(-1);
	}
	s1 = x->result->value.val.strval;
	dsvec_add_ptr(argv, s1);
	argc++;
  }

  dsvec_add_ptr(argv, NULL);

  if ((st = file_func(dsvec_len(argv) - 1, (char **) dsvec_base(argv),
					  &result->value.val.strval)) == -1) {
	if (result->value.val.strval != NULL)
	  seterr_e(e, result, result->value.val.strval);
	else
	  seterr_e(e, result, "Error executing 'file' function");
	return(-1);
  }

  result->value.token = T_STRING;
  if (st == 0)
	result->value.val.strval = "";

  return(0);
}

static int
func_file_owner(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int st;
  char *argv[4], *errmsg, *filename, *ident, *jname, *owner, *sresult;
  extern int file_func(int argc, char **argv, char **result);

  if (arglist != NULL) {
	if (arglist->next != NULL)
	  return(-1);
	filename = arglist->result->value.val.strval;
  }
  else
	filename = var_ns_get_value(e->env->namespaces, "DACS", "FILENAME");

  jname = var_ns_get_value(e->env->namespaces, "Conf", "JURISDICTION_NAME");
  argv[0] = "test";
  argv[1] = "-e";
  argv[2] = filename;
  argv[3] = NULL;

  result->value.token = T_INTEGER;
  result->value.val.intval = 0;

  if ((st = file_func(3, argv, &sresult)) == -1)
	return(-1);

  if (st != 1)
	return(0);
 
  argv[0] = "stat";
  argv[1] = "%U";
  argv[2] = filename;
  argv[3] = NULL;
  if ((st = file_func(3, argv, &owner)) == -1)
	return(-1);
  ident = ds_xprintf("%s:%s", jname, owner);
  if ((st = ue_exec(ident, e->env->credentials, NULL, &errmsg)) == -1) {
	if (errmsg != NULL)
	  seterr_e(e, result, ds_xprintf("Error in 'file_owner' function: %s",
									 errmsg));
	return(-1);
  }

  result->value.token = T_INTEGER;
  result->value.val.intval = st;

  return(0);
}

static int
func_file_group(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int st;
  char *argv[4], *errmsg, *filename, *group, *ident, *jname, *sresult;
  extern int file_func(int argc, char **argv, char **result);

  if (arglist != NULL) {
	if (arglist->next != NULL)
	  return(-1);
	filename = arglist->result->value.val.strval;
  }
  else
	filename = var_ns_get_value(e->env->namespaces, "DACS", "FILENAME");

  jname = var_ns_get_value(e->env->namespaces, "Conf", "JURISDICTION_NAME");
  argv[0] = "test";
  argv[1] = "-e";
  argv[2] = filename;
  argv[3] = NULL;

  result->value.token = T_INTEGER;
  result->value.val.intval = 0;

  if ((st = file_func(3, argv, &sresult)) == -1)
	return(-1);

  if (st != 1)
	return(0);
 
  argv[0] = "stat";
  argv[1] = "%G";
  argv[2] = filename;
  argv[3] = NULL;
  if ((st = file_func(3, argv, &group)) == -1)
	return(-1);
  ident = ds_xprintf("%%%s:%s", jname, group);
  if ((st = ue_exec(ident, e->env->credentials, NULL, &errmsg)) == -1) {
	if (errmsg != NULL)
	  seterr_e(e, result, ds_xprintf("Error in 'file_group' function: %s",
									 errmsg));
	return(-1);
  }

  result->value.token = T_INTEGER;
  result->value.val.intval = st;

  return(0);
}

/*
 * Test where a request comes from (as in Apache allow/deny directives),
 * comparing it (REMOTE_ADDR) against:
 * a (partial) domain name, a full IP address, a partial IP address,
 * a network/netmask pair, or a network/nnn CIDR specification.
 * For compatibility with Apache, the argument "all" (case insensitive)
 * always returns 1.
 */
static int
func_from(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int st;
  char *errmsg, *remote_addr, *remote_host, *s1;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  if (*s1 == '\0') {
	seterr_e(e, result, "Invalid argument to 'from' function");
	return(-1);
  }

  remote_addr = var_ns_get_value(e->env->namespaces, "DACS", "REMOTE_ADDR");
  remote_host = var_ns_get_value(e->env->namespaces, "DACS", "REMOTE_HOST");

  if ((st = is_from_address(s1, remote_addr, remote_host, &errmsg)) == -1) {
	seterr_e(e, result,
			 ds_xprintf("%s while evaluating 'from' function", errmsg));
	return(-1);
  }

  result->value.token = T_INTEGER;
  if (st == 1)
	result->value.val.intval = 1;
  else
	result->value.val.intval = 0;

  return(0);
}

/*
 * Read a file (from the filesystem) or retrieve something from a store.
 * Syntax 1: get(<filename>)
 * Syntax 2: get(<item_type>, <key>)
 *
 * NB: trailing newlines are always removed, which is sometimes a feature
 * and sometimes a bug.
 */
static int
func_get(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *buf, *s1, *s2;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;
  x = x->next;

  if (*s1 == '\0') {
	seterr_e(e, result, "Invalid argument to 'get' function");
	return(-1);
  }

  if (x == NULL) {
	if (load_file(s1, &buf, NULL) == -1) {
	  seterr_e(e, result, ds_xprintf("Could not get file \"%s\"", s1));
	  return(-1);
	}
  }
  else {
	Vfs_handle *h;

	if (x->result->value.token != T_STRING) {
	  seterr_e(e, result, "Invalid key argument to 'get' function");
	  return(-1);
	}
	if (x->next != NULL) {
	  seterr_e(e, result, "Too many arguments to 'get' function");
	  return(-1);
	}
	s2 = x->result->value.val.strval;
	if ((h = vfs_open_item_type(s1)) == NULL) {
	  seterr_e(e, result,
			   ds_xprintf("Could not open vfs for item type \"%s\" in 'get' function", s1));
	  return(-1);
	}
	if (vfs_get(h, s2, (void **) &buf, NULL) == -1) {
	  seterr_e(e, result,
			   ds_xprintf("Could not find item \"%s\" in 'get' function", s2));
	  return(-1);
	}

	if (vfs_close(h) == -1) {
	  seterr_e(e, result,
			   ds_xprintf("Close failed for item type \"%s\" in 'get' function", s1));
	  return(-1);
	}
  }

  result->value.token = T_STRING;
  result->value.val.strval = strtrim(buf, "\n");

  return(0);
}

static int
func_pathname(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1, *s2, *s3, *str;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;
  x = x->next;
  s2 = x->result->value.val.strval;
  x = x->next;
  s3 = x->result->value.val.strval;

  if ((str = directory_name_interpolate(s1, s2, s3)) == NULL) {
	seterr_e(e, result, ds_xprintf("Directory name interpolation failed"));
	return(-1);
  }

  result->value.token = T_STRING;
  result->value.val.strval = str;

  return(0);
}

/*
 * Concatenate all arguments, converting to strings if necessary, and print
 * to the log.  Useful for debugging.
 * The returned value is 1.
 */
static int
func_print(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *str;
  Arglist *x;
  Ds ds;

  ds_init(&ds);
  for (x = arglist; x != NULL; x = x->next) {
	if (force_string(x->result, &str) == -1)
	  return(-1);
	ds_asprintf(&ds, "%s", str);
  }

  log_msg(((Log_level) (LOG_DEBUG_LEVEL | LOG_USER_FLAG), "%s", ds_buf(&ds)));

  result->value.token = T_INTEGER;
  result->value.val.intval = 1;

  return(0);
}

#define PRINTF(DS, FMT, FIELD_WIDTH, PRECISION, FUNC) \
  do { \
    if (FIELD_WIDTH) { \
      if (PRECISION) \
        (void) ds_asprintf(DS, FMT, FIELD_WIDTH, PRECISION, FUNC); \
      else \
        (void) ds_asprintf(DS, FMT, FIELD_WIDTH, FUNC); \
    } \
	else if (PRECISION) \
	  (void) ds_asprintf(DS, FMT, PRECISION, FUNC); \
    else \
	  (void) ds_asprintf(DS, FMT, FUNC); \
  } while (0);

/*
 * Scaled down version of printf(3).
 * XXX Might automatically convert arguments to match format spec.
 */
static int
func_printf(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int cc, field_width, i, precision, saved_ch;
  long lval;
  unsigned long ulval;
  double rval;
  char *fmt, *fmt_start, *format, *strval;
  Arglist *x;
  Ds ds;

  x = arglist;
  format = x->result->value.val.strval;
  x = x->next;

  ds_init(&ds);
  while (1) {
	for (fmt = format; *fmt != '\0'; fmt++) {
	  switch (*fmt) {
	  case '%':
		fmt_start = fmt++;
		if (*fmt == '%') {
		  ds_asprintf(&ds, "%%");
		  break;
		}

		while (strchr("#-+ 0", *fmt) != NULL)
		  fmt++;
		if (*fmt == '*') {
		  if (x->result->value.token != T_INTEGER) {
			seterr_e(e, result, "Invalid field width value in 'printf'");
			return(-1);
		  }
		  field_width = x->result->value.val.intval;
		  x = x->next;
		}
		else
		  field_width = 0;

		while (strchr("*0123456789", *fmt) != NULL)
		  fmt++;
		if (*fmt == '.')
		  fmt++;
		if (*fmt == '*') {
		  if (x->result->value.token != T_INTEGER) {
			seterr_e(e, result, "Invalid precision value in 'printf'");
			return(-1);
		  }
		  precision = x->result->value.val.intval;
		  x = x->next;
		}
		else
		  precision = 0;

		while (strchr("*0123456789", *fmt) != NULL)
		  fmt++;
		if (*fmt == '\0') {
		  seterr_e(e, result, "Missing conversion specifier in 'printf'");
		  return(-1);
		}
		saved_ch = *(fmt + 1);
		*(fmt + 1) = '\0';
		switch (*fmt) {
		case 'c':
		case 's':
		  if (x == NULL)
			strval = "";
		  else {
			if (x->result->value.token != T_STRING) {
			  seterr_e(e, result, "String argument required in 'printf'");
			  return(-1);
			}
			strval = x->result->value.val.strval;
			x = x->next;
		  }
		  PRINTF(&ds, fmt_start, field_width, precision, strval);
		  break;

		case 'd':
		case 'i':
		  if (x == NULL)
			lval = 0;
		  else {
			if (x->result->value.token != T_INTEGER) {
			  seterr_e(e, result, "Integer argument required in 'printf'");
			  return(-1);
			}
			lval = x->result->value.val.intval;
			x = x->next;
		  }
		  PRINTF(&ds, fmt_start, field_width, precision, lval);
		  break;

		case 'o':
		case 'u':
		case 'x':
		case 'X':
		  if (x == NULL)
			ulval = 0;
		  else {
			if (x->result->value.token != T_INTEGER) {
			  seterr_e(e, result, "Integer argument required in 'printf'");
			  return(-1);
			}
			ulval = x->result->value.val.intval;
			x = x->next;
		  }
		  PRINTF(&ds, fmt_start, field_width, precision, ulval);
		  break;

		case 'e':
		case 'E':
		case 'f':
		case 'g':
		case 'G':
		  if (x == NULL)
			rval = 0.0;
		  else {
			if (x->result->value.token != T_REAL) {
			  seterr_e(e, result, "Real argument required in 'printf'");
			  return(-1);
			}
			rval = x->result->value.val.realval;
			x = x->next;
		  }
		  PRINTF(&ds, fmt_start, field_width, precision, rval);
		  break;
		default:
		  return(-1);
		}

		*(fmt + 1) = saved_ch;
		break;

	  case '\\':
		fmt++;
		if (*fmt == '\0')	/* Ends with a backslash?  Ignore it. */
		  break;
		if (is_octdigit((int) *fmt)) {
		  cc = 0;
		  for (i = 0; i < 3; i++) {
			if (!is_octdigit((int) *fmt))
			  break;
			cc = cc * 8 + octdigit_val((int) *fmt);
			fmt++;
		  }
		}
		else {
		  /* Harbison & Steele 2.7.6 */
		  switch ((int) *fmt) {
		  case 'a':
			cc = '\a'; break;
		  case 'b':
			cc = '\b'; break;
		  case 'f':
			cc = '\f'; break;
		  case 'n':
			cc = '\n'; break;
		  case 'r':
			cc = '\r'; break;
		  case 't':
			cc = '\t'; break;
		  case 'v':
			cc = '\v'; break;
		  default:
			cc = (int) *fmt; break;
		  }
		}

		ds_asprintf(&ds, "%c", cc);
		break;

	  default:
		ds_asprintf(&ds, "%c", *fmt);
		break;
	  }
	}

	/* If we've used up both the format and the arguments, we're done. */
	if (x == NULL)
	  break;
  }

  log_msg(((Log_level) (LOG_DEBUG_LEVEL | LOG_USER_FLAG), "%s", ds_buf(&ds)));

  result->value.token = T_INTEGER;
  result->value.val.intval = 1;

  return(0);
}

/*
 * Execute a program with the specified arguments.
 * Return NULL if there's an error, otherwise a string.
 * Usage: exec(<path>[, arg]...)
 */
static int
func_exec(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int read_fd, status;
  char *s1;
  pid_t pid, wpid;
  Arglist *x;
  Ds ds;
  Dsvec *v;
  FILE *fp;

  x = arglist;
  if (x == NULL) {
	seterr_e(e, result, "Invalid argument to 'exec' function");
	return(-1);
  }

  v = dsvec_init(NULL, sizeof(char *));
  for (x = arglist; x != NULL; x = x->next) {
	if (force_string(x->result, &s1) == -1) {
	  seterr_e(e, result, "Invalid argument to 'exec' function");
	  return(-1);
	}
	if (x == arglist && *s1 == '\0') {
	  seterr_e(e, result, "Invalid argument to 'exec' function");
	  return(-1);
	}
	dsvec_add_ptr(v, s1);
  }

  /* XXX Should there be a time limit to wait? */
  dsvec_add_ptr(v, NULL);
  if (filterthru((char **) dsvec_base(v), NULL, &read_fd, NULL, NULL, &pid)
	  == -1) {
	seterr_e(e, result, "Function 'exec' failed");
	return(-1);
  }

  if ((fp = fdopen(read_fd, "r")) == NULL) {
	close(read_fd);
	seterr_e(e, result, "I/O failure in function 'exec'");
	return(-1);
  }

  ds_init(&ds);
  while (ds_agets(&ds, fp) != NULL)
	;
  if (ferror(fp) || !feof(fp)) {
	seterr_e(e, result, "I/O failure while reading in function 'exec'");
	return(-1);
  }
  fclose(fp);

  if ((wpid = waitpid(pid, &status, WNOHANG)) != pid) {
	seterr_e(e, result, "Error during waitpid call in function 'exec'");
	return(-1);
  }

  var_ns_add_key(e->env->namespaces, "DACS",
				 "status", ds_xprintf("%d", WEXITSTATUS(status)));
  result->value.token = T_STRING;
  if (ds_len(&ds) == 0)
	result->value.val.strval = strdup("");
  else
	result->value.val.strval = ds_buf(&ds);

#ifdef NOTDEF
  lastch = result->value.val.strval + ds_len(&ds) - 2;
  if (*lastch == '\n')
	*lastch = '\0';
#endif

  return(0);
}

/*
 * Invoke the specified URL with the given arguments.
 * Return NULL if there's an error, otherwise a string.
 *
 * Usage: http(<url>, <method>[, argname, argvalue]...)
 * If <url> begins with "https", use SSL.
 * The <method> is "GET", "POST", etc. (case insensitive).
 */
static int
func_http(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int n, reply_len, st, status_code;
  char *m, *url, *argname, *argvalue, *reply;
  Arglist *x;
  Dsvec *v;
  Http_params *params;
  Http_method method;

  x = arglist;
  url = x->result->value.val.strval;
  if (*url == '\0') {
	seterr_e(e, result, "Invalid url argument to 'http' function");
	return(-1);
  }
  x = x->next;
  m = x->result->value.val.strval;
  if ((method = http_string_to_method(m)) == HTTP_UNKNOWN_METHOD) {
	seterr_e(e, result, "Unrecognized method argument to 'http' function");
	return(-1);
  }

  v = dsvec_init(NULL, sizeof(Http_params));

  for (n = 0, x = x->next; x != NULL; n++, x = x->next) {
	if (x->result->value.token != T_STRING) {
	  seterr_e(e, result, "Invalid argument to 'http' function");
	  return(-1);
	}
	argname = x->result->value.val.strval;

	x = x->next;
	if (x == NULL) {
	  seterr_e(e, result, "Missing argument to 'http' function");
	  return(-1);
	}
	if (force_string(x->result, &argvalue) == -1) {
	  seterr_e(e, result, "Invalid argument to 'http' function");
	  return(-1);
	}

	params = dsvec_new_obj(v, Http_params *);
	params->name = argname;
	params->value = argvalue;
	params->filename = NULL;
  }

  reply_len = -1;
  st = http_invoke(url, method, HTTP_SSL_URL_SCHEME,
				   n, (Http_params *) dsvec_base(v), NULL, NULL,
				   &reply, &reply_len, &status_code, NULL);

  if (st == -1) {
	if (reply != NULL)
	  seterr_e(e, result,
			 ds_xprintf("Evaluation of 'http' function failed: %s", reply));
	else
	  seterr_e(e, result, "Evaluation of 'http' function failed");
	return(-1);
  }

  result->value.token = T_STRING;
  result->value.val.strval = reply;
  return(0);
}

/*
 * Usage: setvar(<namespace>, <format>, <kwv-sep-char>, <sep-chars>, <string>)
 * Examples:
 * 0. Call setvar("Q", "url", "a=b&c=d&foo=baz")
 *    Result: Q::a=b, Q::c=d, Q::foo=baz
 * 1. Call setvar("Foo", "kwv", <kwv-sep-char>, <sep-chars>, <string>)
 *    setvar("Foo", "kwv", "=", "&", "a=b&c=d&e=f")
 *    Result: True and ${Foo::a}="b", ${Foo::c}="d", ${Foo::e}="f"
 * 2. Call: setvar("Foo", "kwv", "=", " ", "a=\"hi\" c=\"a b\" e=\"\"")
 *    Result: True and ${Foo::a}="hi", ${Foo::c}="a b", ${Foo::e}=""
 * 3. Call: setvar("Foo", "kwv", "=", ";", "a=hi world;c=a b; e=;")
 *    Result: True and ${Foo::a}="hi world", ${Foo::c}="a b", ${Foo::e}=""
 * 4. Call: setvar("Foo", "kwv", "", " ", "a=hi world;c=a b; e=;")
 *    Result: True and ${Foo::1}="a=hi" ${Foo::2}="world;c=a",
 *    ${Foo::3}="b;", ${Foo::4}="e=;"
 * 5. Call: setvar("Foo", "xml", "${prog_result}")
 *    Result: True if no error element and ${Foo::Var1}="Var1 value",
 *    ${Foo:Var2}="Var2 value"
 * 6. Call: setvar("Foo", "xpointer", <xpath-expression>, <string>)
 */
static int
func_setvar(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *format, *str, *ns;
  Kwv *kwv;
  Arglist *x;

  x = arglist;
  ns = x->result->value.val.strval;
  if (*ns == '\0') {
	seterr_e(e, result, "Invalid namespace argument to 'setvar'");
	return(-1);
  }
  x = x->next;
  format = x->result->value.val.strval;
  x = x->next;

  if (strcaseeq(format, "url")) {
	if (x->result->value.token != T_STRING) {
	  seterr_e(e, result, "Invalid string argument to 'setvar'");
	  return(-1);
	}
	str = x->result->value.val.strval;
	kwv = kwv_init(10);
	if (cgiparse(NULL, str, kwv, NULL) == -1) {
	  seterr_e(e, result, "Error parsing string argument to 'setvar'");
	  return(-1);
	}
	kwv_text(stderr, kwv);
	/* ??? */
  }
  else if (strcaseeq(format, "kwv")) {
	char *kwv_sepchar, *sepchars;

	if (x == NULL) {
	  seterr_e(e, result,
			   "Missing kwv-sep-char argument to 'setvar' function");
	  return(-1);
	}
	if (x->result->value.token != T_STRING) {
	  seterr_e(e, result,
			   "Invalid kwv-sep-char argument to 'setvar' function");
	  return(-1);
	}
	kwv_sepchar = x->result->value.val.strval;
	if (strlen(kwv_sepchar) > 1) {
	  seterr_e(e, result,
			   "Invalid kwv-sep-char argument to 'setvar' function");
	  return(-1);
	}

	x = x->next;
	if (x == NULL) {
	  seterr_e(e, result, "Missing sep-char argument to 'setvar' function");
	  return(-1);
	}
	if (x->result->value.token != T_STRING) {
	  seterr_e(e, result, "Invalid sep-char argument to 'setvar' function");
	  return(-1);
	}
	sepchars = x->result->value.val.strval;

	x = x->next;
	if (x == NULL) {
	  seterr_e(e, result, "Missing string argument to 'setvar' function");
	  return(-1);
	}
	if (force_string(x->result, &str) == -1) {
	  seterr_e(e, result, "Invalid string argument to 'setvar' function");
	  return(-1);
	}

	x = x->next;
	if (x != NULL) {
	  seterr_e(e, result, "Too many arguments to 'setvar' function");
	  return(-1);
	}

	/*
	 * If kwv_sepchar is not a null, it is a character that separates
	 * a keyword from its value; otherwise, there is no keyword and we'll
	 * generate them automatically starting with '1'.
	 * If not a null, sepchars is one or more characters, any of which
	 * separate one keyword/value (or just a value) from the next.
	 */
  }
  else if (strcaseeq(format, "xml")) {
	if (x == NULL) {
	  seterr_e(e, result, "Missing string argument to 'setvar' function");
	  return(-1);
	}
	if (force_string(x->result, &str) == -1) {
	  seterr_e(e, result, "Invalid string argument to 'setvar' function");
	  return(-1);
	}

	x = x->next;
	if (x != NULL) {
	  seterr_e(e, result, "Too many arguments to 'setvar' function");
	  return(-1);
	}

	/*
	 * The string argument is expected to be an XML document that
	 * describes a set of variable names and their values.
	 */
	seterr_e(e, result, "Function 'setvar' is not fully implemented");
	return(-1);
  }
  else {
	seterr_e(e, result, "Invalid format argument to 'setvar' function");
	return(-1);
  }

  result->value.token = T_INTEGER;
  result->value.val.intval = 1;
  return(0);
}

/*
 * XXX Given that system(3) is susceptible to a variety of security problems,
 * I'm not sure that this should be implemented... perhaps if it is done
 * carefully and with certain restrictions (e.g., see perlsec(1)).
 */
static int
func_system(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1;
  Arglist *x;

  x = arglist;
  s1 = x->result->value.val.strval;

  seterr_e(e, result, "Attempt to execute unimplemented function: system");
  return(-1);
}

/*
 * Return information about the current time and date.
 * Usage: time(<timearg>[, <timeval>])
 */
static int
func_time(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  char *s1;
  Arglist *x;
  time_t now;
  struct tm *tm;

  x = arglist;
  s1 = x->result->value.val.strval;
  x = x->next;

  if (x != NULL) {
	long v;

	if (force_integer(x->result, &v) == -1) {
	  seterr_e(e, result, "Invalid timeval argument to 'time' function");
	  return(-1);
	}
	now = (time_t) v;
	x = x->next;
  }
  else
	time(&now);

  if (*s1 == '\0' || x != NULL) {
	seterr_e(e, result, "Invalid argument to 'time' function");
	return(-1);
  }

  tm = localtime(&now);

  if (strcaseeq(s1, "now")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = now;
  }
  else if (strcaseeq(s1, "secs") || strcaseeq(s1, "seconds")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_sec;
  }
  else if (strcaseeq(s1, "mins") || strcaseeq(s1, "minutes")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_min;
  }
  else if (strcaseeq(s1, "hour")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_hour;
  }
  else if (strcaseeq(s1, "mday")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_mday;
  }
  else if (strcaseeq(s1, "mon") || strcaseeq(s1, "month")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_mon;
  }
  else if (strcaseeq(s1, "year")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_year + 1900;
  }
  else if (strcaseeq(s1, "wday")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_wday;
  }
  else if (strcaseeq(s1, "yday")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_yday;
  }
  else if (strcaseeq(s1, "isdst")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_isdst;
  }
#if !defined(DACS_OS_CYGWIN) && !defined(DACS_OS_SOLARIS)
  else if (strcaseeq(s1, "zone")) {
	result->value.token = T_STRING;
	if (tm->tm_zone != NULL)
	  result->value.val.strval = (char *) tm->tm_zone;
	else
	  result->value.val.strval = "";
  }
  else if (strcaseeq(s1, "gmtoff")) {
	result->value.token = T_INTEGER;
	result->value.val.intval = tm->tm_gmtoff;
  }
#endif
  else {
	seterr_e(e, result, "Invalid argument to 'time' function");
	return(-1);
  }

  return(0);
}

static int
list_add(char *naming_context, char *name, void ***ptr)
{
  Dsvec *dsv;

  if ((dsv = (Dsvec *) *ptr) == NULL) {
    *ptr = (void **) dsvec_init(NULL, sizeof(char *));
    dsv = (Dsvec *) *ptr;
  }

  dsvec_add_ptr(dsv, strdup(name));

  return(1);
}

/*
 * Virtual filestore access
 * Usage: vfs(<op>, <item_type>[, arg]...)
 */
static int
func_vfs(Lex_state *e, Arglist *arglist, Expr_result *result)
{
  int st;
  char *buf, *item_type, *key, *name, *opname, *s1;
  Arglist *x;
  Ds *err;
  Vfs_conf conf;
  Vfs_directive *vd;
  Vfs_handle *h;
  Vfs_op op;

  x = arglist;
  opname = x->result->value.val.strval;
  if ((op = vfs_lookup_op(opname)) == VFS_UNKNOWN) {
	seterr_e(e, result, "Unrecognized op argument to 'vfs' function");
	return(-1);
  }
  if (op == VFS_OPEN || op == VFS_CLOSE) {
	seterr_e(e, result, "Invalid op argument to 'vfs' function");
	return(-1);
  }
  log_msg((LOG_TRACE_LEVEL, "vfs op=\"%s\"", opname));

  x = x->next;
  if (x == NULL) {
	seterr_e(e, result, "Missing item_type argument to 'vfs' function");
	return(-1);
  }
  if (x->result->value.token != T_STRING) {
	seterr_e(e, result, "Invalid item_type argument to 'vfs' function");
	return(-1);
  }
  name = x->result->value.val.strval;
  x = x->next;

  err = ds_init(NULL);
  key = NULL;
  s1 = NULL;
  if (x != NULL) {
	if (force_string(x->result, &key) == -1) {
	  seterr_e(e, result, "Invalid key argument to 'vfs' function");
	  return(-1);
	}
	x = x->next;

	if (x != NULL) {
	  if (force_string(x->result, &s1) == -1) {
		seterr_e(e, result, "Invalid argument to 'vfs' function");
		return(-1);
	  }
	  x = x->next;

	  if (x != NULL) {
		seterr_e(e, result, "Too many arguments to 'vfs' function");
		return(-1);
	  }
	}
  }

  conf.null_flag = 0;
  conf.create_flag = (op != VFS_EXISTS && op != VFS_DELETE
					  && op != VFS_RENAME);
  vfs_conf(&conf);

  if ((vd = vfs_uri_parse(name, NULL)) != NULL) {
	if ((h = vfs_open_uri(name)) == NULL) {
	  ds_asprintf(err, "Invalid vfs_uri: \"%s\", ", name);
	  ds_asprintf(err, "failed in 'vfs' function");
	  seterr_e(e, result, ds_buf(err));
	  return(-1);
	}
  }
  else {
	item_type = name;
	if ((vd = vfs_lookup_item_type(item_type)) == NULL) {
	  ds_asprintf(err, "VFS directives are missing or invalid for \"%s\", ",
				  item_type);
	  ds_asprintf(err, "failed in 'vfs' function");
	  seterr_e(e, result, ds_buf(err));
	  return(-1);
	}

	if (vfs_open(vd, &h) == -1) {
	  if (op == VFS_EXISTS) {
		result->value.token = T_INTEGER;
		result->value.val.intval = 0;
		return(0);
	  }

	  ds_asprintf(err, "Could not open VFS for item type \"%s\", ", item_type);
	  ds_asprintf(err, "failed in 'vfs' function");
	  seterr_e(e, result, ds_buf(err));
	  return(-1);
	}
  }

  switch (op) {
  case VFS_GET:
	if (s1 != NULL) {
	  ds_asprintf(err,
				  "Too many arguments to 'get' for key \"%s\" ,", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	if (vfs_get(h, key, (void **) &buf, NULL) == -1) {
	  ds_asprintf(err, "Error during 'get' for key \"%s\", ", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	result->value.token = T_STRING;
	result->value.val.strval = strtrim(buf, "\n");
	break;

  case VFS_GETSIZE:
	{
	  size_t length;

	  if (s1 != NULL) {
		ds_asprintf(err, "Too many arguments to 'getsize' for key \"%s\" ,",
					key);
		ds_asprintf(err, "failed in 'vfs' function: %s",
					h->error_msg != NULL ? h->error_msg : "");
		seterr_e(e, result, ds_buf(err));
		vfs_close(h);
		return(-1);
	  }

	  if (vfs_getsize(h, key, &length) == -1) {
		ds_asprintf(err, "Error during 'getsize' for key \"%s\", ", key);
		ds_asprintf(err, "failed in 'vfs' function: %s",
					h->error_msg != NULL ? h->error_msg : "");
		seterr_e(e, result, ds_buf(err));
		vfs_close(h);
		return(-1);
	  }

	  result->value.token = T_INTEGER;
	  result->value.val.intval = length;
	  break;
	}

  case VFS_EXISTS:
	if (s1 != NULL) {
	  ds_asprintf(err,
				  "Too many arguments to 'exists' for key \"%s\" ,", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	if ((st = vfs_exists(h, key)) == -1) {
	  ds_asprintf(err, "Error during 'exists' for key \"%s\", ", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	result->value.token = T_INTEGER;
	result->value.val.intval = (st > 0);
	break;

  case VFS_DELETE:
	if (s1 != NULL) {
	  ds_asprintf(err,
				  "Too many arguments to 'delete' for key \"%s\" ,", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	if (vfs_delete(h, key) == -1) {
	  ds_asprintf(err, "Error during 'delete' for key \"%s\", ", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	result->value.token = T_INTEGER;
	result->value.val.intval = 1;
	break;

  case VFS_CONTROL:
	{
	  char *c_arg;
	  Vfs_control_op c_op;

	  c_arg = NULL;
	  if (strcaseeq(key, "flush"))
		c_op = VFS_FLUSH_DATA;
	  else if (strcaseeq(key, "set_field_sep")) {
		c_op = VFS_SET_FIELD_SEP;
		if (x == NULL || x->result->value.token != T_STRING
			|| x->next != NULL) {
		  return(-1);
		}
		c_arg = x->result->value.val.strval;
		x = x->next;
	  }
	  else {
		ds_asprintf(err, "Unrecognized 'control' operation \"%s\", ", key);
		ds_asprintf(err, "failed in 'vfs' function: %s",
					h->error_msg != NULL ? h->error_msg : "");
		seterr_e(e, result, ds_buf(err));
		vfs_close(h);
		return(-1);
	  }

	  if ((st = vfs_control(h, c_op, s1, c_arg)) == -1) {
		ds_asprintf(err, "Error during 'control', ");
		ds_asprintf(err, "failed in 'vfs' function: %s",
					h->error_msg != NULL ? h->error_msg : "");
		seterr_e(e, result, ds_buf(err));
		vfs_close(h);
		return(-1);
	  }

	  result->value.token = T_INTEGER;
	  result->value.val.intval = st;
	  break;
	}

  case VFS_PUT:
	if (s1 == NULL) {
	  ds_asprintf(err, "Missing value argument for 'put' for key \"%s\" ,",
				  key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	if (vfs_put(h, key, s1, strlen(s1)) == -1) {
		ds_asprintf(err, "Error during 'put' for key \"%s\", ", key);
		ds_asprintf(err, "failed in 'vfs' function: %s",
					h->error_msg != NULL ? h->error_msg : "");
		seterr_e(e, result, ds_buf(err));
		vfs_close(h);
		return(-1);
	}

	result->value.token = T_INTEGER;
	result->value.val.intval = 1;
	break;

  case VFS_RENAME:
	if (key == NULL) {
	  ds_asprintf(err, "Missing oldkey argument for 'rename' - ");
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}
	if (s1 == NULL) {
	  ds_asprintf(err, "Missing newkey argument for 'rename' - ");
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}
	if (vfs_rename(h, key, s1) == -1) {
	  ds_asprintf(err, "Error during 'rename' for oldkey \"%s\", ", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	result->value.token = T_INTEGER;
	result->value.val.intval = 1;
	break;

  case VFS_LIST:
	{
	  int st;
	  Ds ds;
	  Dsvec *names;

	if (key != NULL || s1 != NULL) {
	  ds_asprintf(err, "Too many arguments to 'list' for key \"%s\" ,", key);
	  ds_asprintf(err, "failed in 'vfs' function: %s",
				  h->error_msg != NULL ? h->error_msg : "");
	  seterr_e(e, result, ds_buf(err));
	  vfs_close(h);
	  return(-1);
	}

	  names = NULL;
	  st = vfs_list(h, NULL, NULL, list_add, (void ***) &names);
	  if (st == -1) {
		ds_asprintf(err, "Error during 'list' for key \"%s\", ", key);
		ds_asprintf(err, "failed in 'vfs' function: %s",
					h->error_msg != NULL ? h->error_msg : "");
		seterr_e(e, result, ds_buf(err));
		vfs_close(h);
		return(-1);
	  }

	  if (names == NULL) {
		result->value.token = T_STRING;
		result->value.val.strval = "";
	  }
	  else {
		unsigned int ui;

		ds_init(&ds);
		for (ui = 0; ui < dsvec_len(names); ui++)
		  ds_asprintf(&ds, "%s%s", (ui == 0) ? "" : "\n",
					  (char *) dsvec_ptr_index(names, ui));

		result->value.token = T_STRING;
		result->value.val.strval = ds_buf(&ds);
	  }
	  break;
	}

  default:
	seterr_e(e, result,
			 ds_xprintf("Unimplemented op argument to 'vfs' function"));
	vfs_close(h);
	return(-1);
  }

  if (vfs_close(h) == -1) {
	ds_asprintf(err, "Close failed for item type \"%s\"", s1);
	ds_asprintf(err, "failed in 'vfs' function: %s",
				h->error_msg != NULL ? h->error_msg : "");
	seterr_e(e, result, ds_buf(err));
	return(-1);
  }

  return(0);
}

#ifdef NOTDEF
/*
 * Type conversion: try to convert the argument to a real value.
 */
static int
func_real(Lex_state *e, Arglist *arglist, Expr_result *result)
{
}

/*
 * Type conversion: try to convert the argument to an integer value.
 */
static int
func_integer(Lex_state *e, Arglist *arglist, Expr_result *result)
{
}

/*
 * Type conversion: try to convert the argument to a string.
 */
static int
func_string(Lex_state *e, Arglist *arglist, Expr_result *result)
{
}

/*
 * Type conversion: try to convert the argument to a boolean value.
 */
static int
func_boolean(Lex_state *e, Arglist *arglist, Expr_result *result)
{
}

/*
 * Manage namespaces, such as setting attributes (e.g., read-only), scope,
 * testing existence, deleting, etc.
 */
static int
func_namespace(Lex_state *e, Arglist *arglist, Expr_result *result)
{
}

/*
 * strptime(3) analog.
 * Usage: strptime(<datestr>, <formatstr>, <namespace>)
 */
static int
func_strptime(Lex_state *e, Arglist *arglist, Expr_result *result)
{
}

/*
 * Delays the caller
 * Usage: sleep(<seconds>)
 */
static int
func_sleep(Lex_state *e, Arglist *arglist, Expr_result *result)
{

}

/*
 * Returns a random value
 * Usage: random(<lower-limit>, <upper-limit>)
 * If either argument is a real value, the result is a real, otherwise
 * it is an integer.
 */
static int
func_random(Lex_state *e, Arglist *arglist, Expr_result *result)
{

}

/*
 * Return information about credentials.
 */
static int
func_credentials(Lex_state *e, Arglist *arglist, Expr_result *result)
{

}

/*
 * Return information about the request:
 * the remote IP address, the remote domain name(s), user agent, method.
 */
static int
func_request(Lex_state *e, Arglist *arglist, Expr_result *result)
{

}
#endif

static Func *
lookup_function(char *name)
{
  Func *f;

#ifdef NOTDEF
  if (module_lookup_function(name, NULL, NULL) != -1) {
	log_msg((LOG_TRACE_LEVEL,
			 "Calling dynamically loaded function: %s", name));
	return(NULL);
  }
#endif

  for (f = functions; f->name != NULL; f++) {
	/*
	 * XXX We might support variadic functions by continuing to search
	 * for a "closer" matching function in the list.
	 */
	if (streq(f->name, name))
	  return(f);
  }

  return(NULL);
}

/*
 * Evaluate function calls
 */
static int
eval_function(char *name, Lex_state *e, Arglist *args, Expr_result *result)
{
  int st;
  Func *f;

  if ((f = lookup_function(name)) == NULL) {
	seterr_e(e, result,
			 ds_xprintf("Unrecognized function name: \"%s\"", name));
	return(-1);
  }

  if (validate_function(e, f, args) == -1) {
	seterr_e(e, result, ds_xprintf("Invalid function call: \"%s\"", name));
	return(-1);
  }

  result->err = ACS_EXPR_FALSE;
  result->errmsg = NULL;

  if (!e->do_eval)
	return(0);

  /*
   * XXX Maybe store a copy of the result of any successful function call
   * in ${Expr::result} so that it can be referenced later.
   */
  st = f->func(e, args, result);
  if (st == -1)
	return(-1);

  return(0);
}

/*
 * Evaluate unary operators
 */
static int
eval1(Lex_state *e, Expr_result *result, Token op, Expr_result *r1)
{
  long l1, val;
  double real1;

  if (!e->do_eval)
	return(0);

  val = 0;

  switch (op) {
  case T_UNARY_MINUS:
	if (r1->value.token == T_INTEGER) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = -(r1->value.val.intval);
	}
	else if (r1->value.token == T_REAL) {
	  result->value.token = T_REAL;
	  result->value.val.realval = -(r1->value.val.realval);
	}
	else if (force_integer(r1, &l1) != -1) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = -l1;
	}
	else if (force_real(r1, &real1) != -1) {
	  result->value.token = T_REAL;
	  result->value.val.realval = -real1;
	}
    else
	  seterr_s(e, result, "Invalid unary minus operand");
    break;

  case T_UNARY_PLUS:
	if (r1->value.token == T_INTEGER) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = r1->value.val.intval;
	}
	else if (r1->value.token == T_REAL) {
	  result->value.token = T_REAL;
	  result->value.val.realval = r1->value.val.realval;
	}
	else if (force_integer(r1, &l1) != -1) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = l1;
	}
	else if (force_real(r1, &real1) != -1) {
	  result->value.token = T_REAL;
	  result->value.val.realval = real1;
	}
    else
	  seterr_s(e, result, "Invalid unary plus operand");
    break;

  case T_NOT:
	if (r1->value.token == T_INTEGER) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = (r1->value.val.intval == 0);
	}
	else if (r1->value.token == T_REAL) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = (r1->value.val.realval == 0.0);
	}
	else if (r1->value.token == T_STRING) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = !(r1->value.val.strval[0] != '\0');
	}
	else
	  seterr_s(e, result, "Invalid unary not operand");
	break;

  case T_BITCPL:
	if (force_integer(r1, &l1) != -1) {
	  result->value.token = T_INTEGER;
	  result->value.val.intval = ~l1;
	}
	else {
	  seterr_s(e, result, "Invalid unary operand");
	  return(-1);
	}
	break;

  default:
	seterr_s(e, result, "Unrecognized unary operator");
	break;
  }

  if (result->err < 0)
	return(-1);

  return(0);
}

/*
 * No check for overflow
 */
static long
lpower(long a, long b)
{
  int i;
  long val;

  if (b < 0L)
	return(0L);

  if (b == 0L)
	return(1L);
  val = 1L;
  for (i = 0; i < b; i++)
	val *= a;

  return(val);
}

static int
indexset(char *s1, char *s2)
{
  size_t n;

  n = strcspn(s1, s2);
  if (n == strlen(s1))
	n = 0;
  else
	n++;

  return((int) n);
}

/*
 * Evaluate binary operators
 */
static int
eval2(Lex_state *e, Expr_result *result, Token op, Expr_result *r1,
	  Expr_result *r2)
{
  int i1, i2;
  long l1, l2, val;
  double real1, real2, rval;
  char *s1, *s2;

  if (!e->do_eval)
	return(0);

  val = 0;

  switch (op) {
  case T_OR:
  case T_AND:
	/*
	 * These are actually handled while parsing, because of their special
	 * semantics, but this evaluation code might still be useful.
	 */
	if (force_boolean(r1, &i1) == -1 || force_boolean(r2, &i2) == -1) {
	  seterr_s(e, result, "Invalid operand");
	  return(-1);
	}

	if (op == T_OR) {
	  if (i1 || i2) {
		result->value.val.intval = 1L;
		result->value.token = T_INTEGER;
	  }
	  else {
		result->value.val.intval = 0L;
		result->value.token = T_INTEGER;
	  }
	}
	else if (op == T_AND) {
	  if (i1 && i2) {
		result->value.val.intval = 1L;
		result->value.token = T_INTEGER;
	  }
	  else {
		result->value.val.intval = 0L;
		result->value.token = T_INTEGER;
	  }
	}
	else {
	  seterr_e(e, result, "Code botch");
	  return(-1);
	}

	break;

  case T_LT:
  case T_LE:
  case T_EQ:
  case T_NE:
  case T_GE:
  case T_GT:
  case T_LT_I:
  case T_LE_I:
  case T_EQ_I:
  case T_NE_I:
  case T_GE_I:
  case T_GT_I:
	if (force_integer(r1, &l1) != -1 && force_integer(r2, &l2) != -1) {
	  ;
	}
	else if (force_real(r1, &real1) != -1 && force_real(r2, &real2) != -1) {
	  l2 = 0L;
	  if (op == T_LT || op == T_LT_I)
		l1 = (real1 < real2) ? -1 : 1;
	  else if (op == T_LE || op == T_LE_I)
		l1 = (real1 <= real2) ? -1 : 1;
	  else if (op == T_EQ || op == T_EQ_I)
		l1 = (real1 == real2) ? 0 : 1;
	  else if (op == T_NE || op == T_NE_I)
		l1 = (real1 != real2) ? -1 : 0;
	  else if (op == T_GE || op == T_GE_I)
		l1 = (real1 >= real2) ? 1 : 0;
	  else if (op == T_GT || op == T_GT_I)
		l1 = (real1 > real2) ? 1 : 0;
	  else {
		seterr_e(e, result, "Code botch");
		return(-1);
	  }
	}
	else if (force_string(r1, &s1) != -1 && force_string(r2, &s2) != -1) {
	  /* Do a string comparison */
	  if (op == T_LT_I || op == T_LE_I || op == T_EQ_I || op == T_NE_I
		  || op == T_GE_I || op == T_GT_I)
		l1 = strcasecmp(s1, s2);
	  else
		l1 = strcmp(s1, s2);
	  l2 = 0L;
	}
	else {
	  seterr_s(e, result, "Invalid binary operand");
	  return(-1);
	}

	switch (op) {
	case T_LT:
	case T_LT_I:
	  val = (l1 < l2);
	  break;
	case T_LE:
	case T_LE_I:
	  val = (l1 <= l2);
	  break;
	case T_EQ:
	case T_EQ_I:
	  val = (l1 == l2);
	  break;
	case T_NE:
	case T_NE_I:
	  val = (l1 != l2);
	  break;
	case T_GE:
	case T_GE_I:
	  val = (l1 >= l2);
	  break;
	case T_GT:
	case T_GT_I:
	  val = (l1 > l2);
	  break;
	default:
	  seterr_e(e, result, "Code botch");
	  return(-1);
	}

	result->value.token = T_INTEGER;
	result->value.val.intval = val;
	break;

  case T_BITOR:
  case T_BITXOR:
  case T_BITAND:
  case T_BITSHL:
  case T_BITSHR:
	if (force_integer(r1, &l1) != -1 && force_integer(r2, &l2) != -1) {
	  if (op == T_BITOR)
		val = l1 | l2;
	  else if (op == T_BITXOR)
		val = l1 ^ l2;
	  else if (op == T_BITAND)
		val = l1 & l2;
	  else if (op == T_BITSHL)
		val = l1 << l2;
	  else if (op == T_BITSHR)
		val = l1 >> l2;
	  else {
		seterr_e(e, result, "Code botch");
		return(-1);
	  }
	  result->value.token = T_INTEGER;
	  result->value.val.intval = val;
	}
	else {
	  seterr_s(e, result, "Invalid binary operand");
	  return(-1);
	}
	break;

  case T_PLUS:
  case T_MINUS:
  case T_TIMES:
  case T_DIV:
	if (force_integer(r1, &l1) != -1 && force_integer(r2, &l2) != -1) {
	  if (op == T_PLUS)
		val = l1 + l2;
	  else if (op == T_MINUS)
		val = l1 - l2;
	  else if (op == T_TIMES)
		val = l1 * l2;
	  else if (op == T_DIV) {
		if (l2 == 0) {
		  seterr_e(e, result, "Attempt to divide by zero");
		  return(-1);
		}
		val = l1 / l2;
	  }
	  else {
		seterr_e(e, result, "Code botch");
		return(-1);
	  }

	  result->value.token = T_INTEGER;
	  result->value.val.intval = val;
	  break;
	}
	else if (force_real(r1, &real1) != -1 && force_real(r2, &real2) != -1) {
	  if (op == T_PLUS)
		rval = real1 + real2;
	  else if (op == T_MINUS)
		rval = real1 - real2;
	  else if (op == T_TIMES)
		rval = real1 * real2;
	  else if (op == T_DIV) {
		if (real2 == 0.0) {
		  seterr_e(e, result, "Attempt to divide by zero");
		  return(-1);
		}
		rval = real1 / real2;
	  }
	  else {
		seterr_e(e, result, "Code botch");
		return(-1);
	  }

	  result->value.token = T_REAL;
	  result->value.val.realval = rval;
	  break;
	}
	else {
	  seterr_s(e, result, "Invalid binary operand");
	  return(-1);
	}
	break;

  case T_DOT:
	/* String concatenation */
	if (force_string(r1, &s1) == -1 || force_string(r2, &s2) == -1) {
	  seterr_s(e, result, "Invalid binary operand");
	  return(-1);
	}

	result->value.token = T_STRING;
	result->value.val.strval = ds_xprintf("%s%s", s1, s2);
	break;

  case T_EXP:
	if (force_integer(r1, &l1) != -1 && force_integer(r2, &l2) != -1) {
	  if (l2 < 0) {
		/* Can't raise to a negative power */
		seterr_e(e, result, "Attempt to raise to a negative power");
		return(-1);
	  }
	  val = lpower(l1, l2);
	  result->value.token = T_INTEGER;
	  result->value.val.intval = val;
	}
	else if (force_real(r1, &real1) != -1 && force_real(r2, &real2) != -1) {
	  rval = pow(real1, real2);
	  if (errno == ERANGE || errno == EDOM) {
		seterr_e(e, result, "Error or overflow in pow()");
		return(-1);
	  }
	  result->value.token = T_REAL;
	  result->value.val.realval = rval;
	}
	else {
	  seterr_s(e, result, "Invalid binary operand");
	  return(-1);
	}
	break;

  case T_MOD:
	if (force_integer(r1, &l1) != -1 && force_integer(r2, &l2) != -1) {
	  if (l2 == 0L) {
		seterr_e(e, result, "Attempt to divide by zero");
		return(-1);
	  }
	  val = l1 % l2;
	}
	else {
	  seterr_s(e, result, "Invalid binary operand");
	  return(-1);
	}

	result->value.token = T_INTEGER;
	result->value.val.intval = val;
	break;

  default:
	seterr_s(e, result, "Unrecognized binary operator");
	return(-1);
  }

  return(0);
}

#ifdef NOTDEF
/*
 * Evaluate ternary operators
 */
static int
eval3(Lex_state *e, Expr_result *result, Token op, Expr_result *r1,
	  Expr_result *r2, Expr_result *r3)
{
  int rc1, rc2;
  Expr_result res2, res3;

  if (!e->do_eval)
	return(0);

  if (value_init(&res2, result))
	return(-1);
  if (value_init(&res3, result))
	return(-1);

  rc1 = rc2 = 0;

  switch(op) {
  default:
	/* Unknown ternary op */
	seterr_s(e, result, "Unrecognized ternary operator");
	return(-1);
  }

  return(0);
}
#endif

#ifdef PROG

#include "dacs.h"

static void process_commands(FILE *fp);

static void
dacs_usage(char *name)
{

  fprintf(stderr, "Usage: %s %s [-env] [-ef file] [[-e] expr]\n",
		  name, standard_command_line_usage);
  fprintf(stderr, "-env initializes DACS namespace from the environment\n");
  fprintf(stderr, "If no expr is given, interactive mode is used\n");
  fprintf(stderr, "Type '?' for help\n");
  exit(2);
}

static void
show_help(char *str)
{
  int i;
  char *p;
  
  for (p = str; *p == '\t' || *p == ' '; p++)
	;

  if (strcaseprefix(p, "func")) {
	for (i = 1; functions[i - 1].name != NULL; i++) {
	  if ((i % 8) == 0)
		printf(",\n");
	  else if (i != 1)
		printf(", ");
	  printf("%s", functions[i - 1].name);
	}
	printf("\n");
  }
  else {
	printf("clear             - clears all DACS:: variables\n");
	printf("set               - displays all DACS:: variables\n");
	printf("set VAR=\"VALUE\"   - sets DACS:: variable VAR to VALUE\n");
	printf("source filename   - execute commands from a file\n");
	printf("so filename       - same as 'source' command\n");
	printf("<VAR              - evaluate the expression in variable VAR\n");
	printf("quit or exit      - terminates\n");
	printf("help or ?         - this help message\n");
	printf("help func         - list functions\n");
	printf("Anything else is an expression to be evaluated\n");
	printf("A line may be continued by ending it with a backslash\n");
  }

}

static Acs_environment env;

/*
 * Read an expression test case from FILENAME (or stdin, if NULL).
 * A test case consists of a section containing one or more options,
 * each one a single-line comment, followed by the expression to evaluate:
 *
 *    {<whitespace>* "//" <whitespace>* <option> ":" <value> <end-of-line>}*
 * As a special case, lines like the following are ignored:
 *    <whitespace>* "///" .* <end-of-line>
 * Options control how the test is performed and specify the expected result.
 *
 * Options:
 * expect:<regex> or expect-regex:<regex>
 *   - result string must match <regex>
 * expect-exact:<string>
 *   - result string must match <string> exactly
 * expect-code:<code>
 *   - result code must be <code> (0=true, 1=false, 2=error)
 * expect-type:<type>
 *   - result type must be <type> (integer, real, string, literal, or undef)
 * show-result:<y/n> (yes or no)
 *   - if "yes", show the result string
 */
static int
run_test_case(char *filename)
{
  int expect_code, rc, rs, show;
  char *e, *errmsg, *expr, *file, *fname, *option, *p, *str, *test, *value;
  char *expect_exact, *expect_regex, *expect_type;
  char errbuf[100];
  regex_t preg;
  regmatch_t pmatch[1];
  Acs_expr_result st;
  Ds *ds;
  Expr_result *result;

  if (streq(filename, "-")) {
	file = NULL;
	fname = "<stdin>";
  }
  else
	file = fname = filename;

  errmsg = NULL;
  if (load_file(file, &test, NULL) == -1) {
	errmsg = "Error loading test case file";
	goto fail;
  }

  p = test;
  expr = NULL;
  expect_exact = expect_regex = expect_type = NULL;
  expect_code = -1;
  show = 0;

  while (1) {
	while (*p == ' ' || *p == '\t' || *p == '\n')
	  p++;
	if (*p != '/' || *(p + 1) != '/')
	  break;

	p += 2;

	if (*p == '/') {
	  p = strchr(p + 1, (int) '\n');
	  if (p == NULL) {
		errmsg = "No newline terminates option comment";
		goto fail;
	  }
	  p++;
	  continue;
	}

	while (*p == ' ' || *p == '\t')
	  p++;

	option = p;
	if ((p = strchr(option, (int) ':')) == NULL) {
	  errmsg = "Invalid option specification";
	  goto fail;
	}
	*p++ = '\0';

	/* No whitespace is allowed here. */
	value = p;
	if ((p = strchr(value, (int) '\n')) == NULL) {
	  errmsg = "No newline found";
	  goto fail;
	}
	*p++ = '\0';
	if (verbose_level)
	  fprintf(stderr, "Option: \"%s\", Value: \"%s\"\n", option, value);

	if (streq(option, "expect-exact"))
	  expect_exact = value;
	else if (streq(option, "expect") || streq(option, "expect-regex")) {
	  expect_regex = value;
	  if ((rs = regcomp(&preg, expect_regex, REG_EXTENDED)) != 0) {
		regerror(rs, &preg, errbuf, sizeof(errbuf));
		errmsg = ds_xprintf("Invalid regular expression: \"%s\": %s",
							expect_regex, errbuf);
		goto fail;
	  }
	}
	else if (streq(option, "expect-code"))
	  expect_code = atoi(value);
	else if (streq(option, "expect-type"))
	  expect_type = value;
	else if (streq(option, "show-result")) {
	  if (strcaseeq(value, "yes"))
		show = 1;
	  else if (strcaseeq(value, "no"))
		show = 0;
	  else {
		errmsg = "Unrecognized show option value";
		goto fail;
	  }
	}
	else {
	  errmsg = "Unrecognized option";
	  goto fail;
	}
  }

  expr = p;

  e = expr;
  if ((ds = acs_elide_comments(expr)) != NULL)
	e = ds_buf(ds);
  if (!quiet_flag)
	fprintf(stderr, "%s: ", fname);
  result = init_expr_result(NULL);
  st = acs_expr(e, &env, result);

  if (st == ACS_EXPR_TRUE)
	rc = 0;
  else if (st == ACS_EXPR_FALSE)
	rc = 1;
  else
	rc = 2;

  if (expect_code != -1 && rc != expect_code) {
	errmsg = ds_xprintf("failed on expect_code: expect %d, got %d",
						expect_code, rc);
	goto fail;
  }

  if (expect_type != NULL) {
	Lex_token *t;

	if ((t = lookup_token(result->value.token)) == NULL) {
	  errmsg = "Internal error: token lookup failed";
	  goto fail;
	}
	if (strcasesuffix(t->token_name, strlen(t->token_name),expect_type)
		== NULL) {
	  errmsg = ds_xprintf("failed on expect_type: expect %s, got %s",
						  expect_type, t->token_name);
	  goto fail;
	}
  }

  if (st != ACS_EXPR_TRUE && st != ACS_EXPR_FALSE)
	return(0);
  
  str = acs_format_result(result);

  if (expect_exact != NULL) {
	if (!streq(str, expect_exact)) {
	  errmsg = ds_xprintf("failed on expect_exact: expect \"%s\", got \"%s\"",
						  expect_exact, str);
	  goto fail;
	}
  }

  if (expect_regex != NULL) {
	rs = regexec(&preg, str, 1, pmatch, 0);
	if (rs != 0) {
      if (rs != REG_NOMATCH) {
		regerror(rs, &preg, errbuf, sizeof(errbuf));
		errmsg = ds_xprintf("Regular expression error: \"%s\": %s",
							expect_regex, errbuf);
		goto fail;
	  }
	  errmsg = ds_xprintf("failed on expect_regex: expect \"%s\", got \"%s\"",
						  expect_regex, str);
	  goto fail;
	}
  }

  if (show) {
	if (result->value.token == T_STRING)
	  printf("\"%s\"\n", str);
	else
	  printf("%s\n", str);
  }

  if (!quiet_flag)
	fprintf(stderr, "ok\n");

  return(0);

 fail:

  if (errmsg != NULL) {
	if (quiet_flag)
	  fprintf(stderr, "%s: ", fname);
	fprintf(stderr, "%s\n", errmsg);
  }

  return(-1);
}

static int
eval_command(char *buf)
{
  Kwv *kwv;

  if ((kwv = var_ns_lookup_kwv(env.namespaces, "DACS")) == NULL) {
	Var_ns *v;

	v = var_ns_new(&env.namespaces, "DACS", NULL);
	kwv = v->kwv;
  }

  if (streq(buf, "clear")) {
	var_ns_delete(&env.namespaces, "DACS");
  }
  else if (streq(buf, "help") || strprefix(buf, "help "))
	show_help(buf + 4);
  else if (streq(buf, "?") || strprefix(buf, "? "))
	show_help(buf + 1);
  else if (streq(buf, "quit") || streq(buf, "exit"))
	return(0);
  else if (streq(buf, "set")) {
	char *str;

	if ((str = kwv_buf(kwv, '=', '\"')) != NULL)
	  printf("%s", str);
  }
  else if (strneq(buf, "set ", 4)) {
	Kwv_conf kwv_conf = { "=", NULL, " ", 0, NULL, 10, NULL, NULL };

	kwv_make_add(kwv, buf + 4, &kwv_conf);
  }
  else if (strneq(buf, "source ", 7) || strneq(buf, "so ", 3)) {
	char *filename, *p;
	FILE *fp;

	for (p = buf; *p != ' '; p++)
	  ;
	while (*p == ' ')
	  p++;
	filename = p;
	if ((fp = fopen(filename, "r")) == NULL) {
	  fprintf(stderr, "Couldn't open %s\n", filename);
	  return(1);
	}
	process_commands(fp);
	fclose(fp);
  }
  else if (streq(buf, "trace")) {
	env.trace_level = 1;
  }
  else if (strneq(buf, "trace ", 6)) {
	if (streq(buf + 6, "on"))
	  env.trace_level = 1;
	else if (streq(buf + 6, "off"))
	  env.trace_level = 0;
	else
	  printf("?\n");
  }
  else {
	char *expr, *str;
	Acs_expr_result st;
	Expr_result *result;

	if (buf[0] == '<') {
	  if ((expr = kwv_lookup_value(kwv, buf + 1)) == NULL)
		expr = buf;
	}
	else if (buf[0] == '\0')
	  return(1);
	else
	  expr = buf;

	env.do_eval = 1;
	result = init_expr_result(NULL);
	st = acs_expr(expr, &env, result);
	if (st == ACS_EXPR_SYNTAX_ERROR)
	  printf("Syntax error\n");
	else if (st == ACS_EXPR_LEXICAL_ERROR)
	  printf("Lexical error\n");
	else if (st == ACS_EXPR_EVAL_ERROR)
	  printf("Evaluation error\n");
	else {
	  str = acs_format_result(result);
	  if (result->value.token == T_STRING || result->value.token == T_LITERAL)
		fprintf(stderr, "\"%s\"\n", str);
	  else
		fprintf(stderr, "%s\n", str);
	}
  }

  return(1);
}

static void
process_commands(FILE *fp)
{
  char *cmd;
  Ds *ds;

  ds = ds_init(NULL);
  ds->delnl_flag = 1;
  ds->escnl_flag = 1;

  while (1) {
#ifdef HAVE_READLINE
	if ((cmd = ds_readline(ds, "> ", ">> ")) == NULL)
	  break;
#else
	if (fp == stdin)
	  printf("> ");
	if ((cmd = ds_gets(ds, fp)) == NULL) {
	  if (feof(fp) || ferror(fp))
		break;
	  continue;
	}
#endif
	if (eval_command(cmd) == 0)
	  break;
  }

}

/*
 * For testing and debugging purposes.
 */
#ifdef BUILD_BUNDLE
int
dacsexpr_main(int argc, char **argv)
#else
int
main(int argc, char **argv)
#endif
{
  int i, rc;
  char *errmsg, *expr;
  Acs_expr_result st;
  Expr_result result;
  Var_ns *env_ns;

  errmsg = NULL;
  if (dacs_init(DACS_STANDALONE, &argc, &argv, NULL, &errmsg) == -1) {
  fail:
	if (errmsg != NULL)
	  fprintf(stderr, "dacsexpr: %s\n", errmsg);
	dacs_usage(argv[0]);
    /*NOTREACHED*/
  }

  if (!quiet_flag && !dacs_saw_command_line_log_level) {
	/* Override the config file default */
	log_set_level(NULL, LOG_WARN_LEVEL);
  }

  env.url = NULL;
  env.namespaces = NULL;
  env.do_eval = 1;
  env.acl_rule = NULL;
  env.notices = NULL;
  env.redirect = NULL;
  if (dacs_conf != NULL && dacs_conf->conf_var_ns != NULL)
	var_ns_add(&env.namespaces, dacs_conf->conf_var_ns);
  env.trace_level = trace_level;
  env.credentials = NULL;

  env_ns = NULL;
  expr = NULL;
  for (i = 1; i < argc; i++) {
	if (streq(argv[i], "-env")) {
	  if (env_ns == NULL && (env_ns = var_ns_from_env("Env")) != NULL)
		var_ns_add(&env.namespaces, env_ns);
	}
	else if (streq(argv[i], "-e")) {
	  if (++i == argc) {
		errmsg = "Expression is missing";
		goto fail;
	  }
	  if (expr != NULL) {
		errmsg = "Only one expression is allowed";
		goto fail;
	  }
	  expr = argv[i];
	}
	else if (streq(argv[i], "-ef")) {
	  char *filename;

	  if (++i == argc) {
		errmsg = "Expression filename is missing";
		goto fail;
	  }
	  if (expr != NULL || i != (argc - 1)) {
		errmsg = "Only one expression is allowed";
		goto fail;
	  }
	  if (streq(argv[i], "-"))
		filename = NULL;
	  else
		filename = argv[i];
	  if (load_file(filename, &expr, NULL) == -1) {
		errmsg = "Error loading expression file";
		goto fail;
	  }
	}
	else if (streq(argv[i], "-et")) {
	  if (++i == argc) {
		errmsg = "Expression test filename is missing";
		goto fail;
	  }
	  if (expr != NULL || i != (argc - 1)) {
		errmsg = "Only one expression is allowed";
		goto fail;
	  }
	  rc = run_test_case(argv[i]);
	  exit(rc == 0 ? 0 : 1);
	}
	else
	  break;
  }

  if (i == (argc - 1)) {
	if (expr != NULL) {
	  errmsg = "Only one expression is allowed";
	  goto fail;
	}
	expr = argv[i];
  }
  else if (i != argc) {
	dacs_usage(argv[0]);
    /*NOTREACHED*/
  }

  if (expr != NULL) {
	char *e, *str;
	Ds *ds;

	e = expr;
	if ((ds = acs_elide_comments(expr)) != NULL)
	  e = ds_buf(ds);
	st = acs_expr(e, &env, &result);

	if (st == ACS_EXPR_TRUE)
	  rc = 0;
	else if (st == ACS_EXPR_FALSE)
	  rc = 1;
	else
	  rc = 2;

	if (!quiet_flag) {
	  if (st == ACS_EXPR_SYNTAX_ERROR)
		fprintf(stderr, "Syntax error\n");
	  else if (st == ACS_EXPR_EVAL_ERROR)
		fprintf(stderr, "Evaluation error\n");
	  else {
		str = acs_format_result(&result);
		if (result.value.token == T_STRING)
		  printf("\"%s\"\n", str);
		else
		  printf("%s\n", str);
	  }
	}

	exit(rc);
  }

  process_commands(stdin);

  exit(0);
}
#endif
